# coding=utf-8
# pylint: skip-file
# FIXME: pylint the shit out of this
import calendar
import typing
from itertools import chain
from time import gmtime, strftime

from emiz import MAIN_LOGGER
from emiz.validator import VALID_BOOL, VALID_FLOAT, VALID_INT, VALID_POSITIVE_INT, VALID_STR, Validator

EPOCH_DELTA = 1306886400

LOGGER = MAIN_LOGGER.getChild(__name__)

validator_group_or_unit_name = Validator(_type=str, _regex=r'[a-zA-Z0-9\_\-\#]+',
                                         exc=ValueError, logger=LOGGER)


class BaseMissionObject:
    def __init__(self, mission_dict: dict, l10n: dict):
        super().__init__()

        if not isinstance(mission_dict, dict):
            raise TypeError('mission_dict should be an dict, got: {}'.format(type(mission_dict)))

        if not isinstance(l10n, dict):
            raise TypeError('l10n should be an dict, got: {}'.format(type(l10n)))

        self.d = mission_dict
        self.l10n = l10n

        self.weather = None
        self._blue_coa = None
        self._red_coa = None
        self.ground_control = None

        self._countries_by_name = {}
        self._countries_by_id = {}

    def get_country_by_name(self, country_name):
        VALID_STR.validate(country_name, 'get_country_by_name', exc=ValueError)
        if country_name not in self._countries_by_name.keys():
            for country in self.countries:
                assert isinstance(country, Country)
                if country.country_name == country_name:
                    self._countries_by_name[country_name] = country
                    return country
            raise ValueError(country_name)
        else:
            return self._countries_by_name[country_name]

    def get_country_by_id(self, country_id):
        VALID_POSITIVE_INT.validate(country_id, 'get_country_by_id')
        if country_id not in self._countries_by_id.keys():
            for country in self.countries:
                assert isinstance(country, Country)
                if country.country_id == country_id:
                    self._countries_by_id[country_id] = country
                    return country
            raise ValueError(country_id)
        else:
            return self._countries_by_id[country_id]

    def get_groups_from_category(self, category):
        Mission.validator_group_category.validate(category, 'get_groups_from_category')
        for group in self.groups:
            if group.group_category == category:
                yield group

    def get_units_from_category(self, category):
        Mission.validator_group_category.validate(category, 'get_units_from_category')
        for unit in self.units:
            if unit.group_category == category:
                yield unit

    def get_group_by_id(self, group_id):
        VALID_POSITIVE_INT.validate(group_id, 'get_group_by_id', exc=ValueError)
        for group in self.groups:
            assert isinstance(group, Group)
            if group.group_id == group_id:
                return group
        return None

    def get_clients_groups(self) -> typing.Generator['Group', None, None]:
        for group in self.groups:
            assert isinstance(group, Group)
            if group.group_is_client_group:
                yield group

    def get_group_by_name(self, group_name):
        VALID_STR.validate(group_name, 'get_group_by_name')
        for group in self.groups:
            assert isinstance(group, Group)
            if group.group_name == group_name:
                return group
        return None

    def get_unit_by_name(self, unit_name):
        VALID_STR.validate(unit_name, 'get_unit_by_name')
        for unit in self.units:
            assert isinstance(unit, BaseUnit)
            if unit.unit_name == unit_name:
                return unit
        return None

    def get_unit_by_id(self, unit_id):
        VALID_POSITIVE_INT.validate(unit_id, 'get_unit_by_id')
        for unit in self.units:
            assert isinstance(unit, BaseUnit)
            if unit.unit_id == unit_id:
                return unit
        return None

    @property
    def units(self):
        for group in self.groups:
            for unit in group.units:
                assert isinstance(unit, BaseUnit)
                yield unit

    @property
    def groups(self):
        for country in self.countries:
            for group in country.groups:
                assert isinstance(group, Group)
                yield group

    @property
    def next_group_id(self):
        ids = set()
        for group in chain(self._blue_coa.groups, self._red_coa.groups):
            assert isinstance(group, Group)
            id_ = group.group_id
            if id_ in ids:
                raise IndexError(group.group_name)
            ids.add(id_)
        return max(ids) + 1

    @property
    def next_unit_id(self):
        ids = set()
        for unit in chain(self._blue_coa.units, self._red_coa.units):
            assert isinstance(unit, BaseUnit)
            id_ = unit.unit_id
            if id_ in ids:
                raise IndexError(unit.unit_name)
            ids.add(id_)
        return max(ids) + 1

    @property
    def coalitions(self):
        for coalition in [self._blue_coa, self._red_coa]:
            assert isinstance(coalition, Coalition)
            yield coalition

    @property
    def countries(self):
        for coalition in self.coalitions:
            for country in coalition.countries:
                assert isinstance(country, Country)
                yield country

    @property
    def _section_date(self):
        return self.d['date']

    @property
    def day(self):
        return self._section_date['Day']

    @day.setter
    def day(self, value):
        Mission.validator_day.validate(value, 'day')
        # Get the last day of the current month
        value = min(value, calendar.monthrange(self.year, self.month)[1])
        self._section_date['Day'] = value

    @property
    def month(self):
        return self._section_date['Month']

    @month.setter
    def month(self, value):
        Mission.validator_month.validate(value, 'month')
        self._section_date['Month'] = value
        self.day = self.day  # update day (for max value)

    @property
    def year(self):
        return self._section_date['Year']

    @year.setter
    def year(self, value):
        Mission.validator_year.validate(value, 'year')
        self._section_date['Year'] = value
        self.day = self.day  # update day (for max value)

    @property
    def mission_start_time(self):
        return self.d['start_time']

    @mission_start_time.setter
    def mission_start_time(self, value):
        Mission.validator_start_time.validate(value, 'start_time')
        self.d['start_time'] = value

    @staticmethod
    def _start_time_as_string(start_time):
        return strftime('%H:%M:%S', gmtime(start_time))

    @staticmethod
    def _start_date_as_string(day, month, year):
        return strftime(f'{day:02}/{month:02}/{year}')

    def _start_datetime_as_string(self, startime):
        return f'{self.mission_start_date_as_string} {self._start_time_as_string(startime)}'

    @property
    def mission_start_time_as_string(self):
        return self._start_time_as_string(self.mission_start_time)

    @property
    def mission_start_date_as_string(self):
        return self._start_date_as_string(self.day, self.month, self.year)

    @property
    def mission_start_datetime_as_string(self):
        return f'{self.mission_start_date_as_string} {self.mission_start_time_as_string}'

    @property
    def _sortie_name_key(self):
        return self.d['sortie']

    @property
    def sortie_name(self):
        return self.l10n[self._sortie_name_key]

    @sortie_name.setter
    def sortie_name(self, value):
        VALID_STR.validate(value, 'sortie name')
        self.l10n[self._sortie_name_key] = value


class Mission(BaseMissionObject):
    validator_start_time = Validator(
        _type=int,
        _min=0,
        _max=86399,
        exc=ValueError,
        logger=LOGGER
    )
    validator_year = Validator(
        _type=int,
        _min=1900,
        _max=2100,
        exc=ValueError,
        logger=LOGGER
    )
    validator_month = Validator(
        _type=int,
        _min=1,
        _max=12,
        exc=ValueError,
        logger=LOGGER
    )
    validator_day = Validator(
        _type=int,
        _min=1,
        _max=31,
        exc=ValueError,
        logger=LOGGER
    )
    validator_heading = Validator(
        _type=int,
        _min=0,
        _max=359,
        exc=ValueError,
        logger=LOGGER
    )
    validator_group_category = Validator(
        _type=str,
        _in_list=['helicopter', 'ship', 'plane', 'vehicle'],
        exc=ValueError,
        logger=LOGGER)
    valid_group_categories = ('helicopter', 'plane', 'ship', 'vehicle')

    def __init__(self, mission_dict, l10n):
        super().__init__(mission_dict, l10n)
        self.weather = Weather(self.d, l10n)
        self._blue_coa = Coalition(self.d, l10n, 'blue')
        self._red_coa = Coalition(self.d, l10n, 'red')
        self.ground_control = GroundControl(self.d, l10n)

    def __repr__(self):
        return 'Mission({})'.format(self.d)

    @property
    def blue_coa(self) -> 'Coalition':
        return self._blue_coa

    @property
    def red_coa(self) -> 'Coalition':
        return self._red_coa

    def farps(self) -> typing.Generator['Static', None, None]:
        for coa in [self._blue_coa, self._red_coa]:
            for farp in coa.farps:
                yield farp


# noinspection PyProtectedMember
class Coalition(BaseMissionObject):
    def __init__(self, mission_dict, ln10, coa_color):
        super().__init__(mission_dict, ln10)
        self.coa_color = coa_color
        self._countries = {}

    def __repr__(self):
        return 'Coalition({}, {})'.format(self._section_coalition, self.coa_color)

    def __eq__(self, other):
        if not isinstance(other, Coalition):
            raise ValueError('"other" must be an Coalition instance; got: {}'.format(type(other)))
        return self._section_coalition == other._section_coalition

    @property
    def _section_coalition(self):
        return self.d['coalition'][self.coa_color]

    @property
    def _section_bullseye(self):
        return self._section_coalition['bullseye']

    @property
    def bullseye_x(self):
        return self._section_bullseye['x']

    @property
    def bullseye_y(self):
        return self._section_bullseye['y']

    @property
    def bullseye_position(self):
        return self.bullseye_x, self.bullseye_y

    @property
    def _section_nav_points(self):
        return self._section_coalition['nav_points']

    @property
    def coalition_name(self):
        return self._section_coalition['name']

    @property
    def _section_country(self):
        return self._section_coalition['country']

    @property
    def countries(self) -> typing.Generator['Country', None, None]:
        for k in self._section_country:
            if k not in self._countries.keys():
                country = Country(self.d, self.l10n, self.coa_color, k)
                self._countries[k] = country
                self._countries_by_id[country.country_id] = country
                self._countries_by_name[country.country_name] = country
            yield self._countries[k]

    def get_country_by_name(self, country_name) -> 'Country':
        VALID_STR.validate(country_name, 'get_country_by_name', exc=ValueError)
        if country_name not in self._countries_by_name.keys():
            for country in self.countries:
                assert isinstance(country, Country)
                if country.country_name == country_name:
                    return country
            raise ValueError(country_name)
        else:
            return self._countries_by_name[country_name]

    def get_country_by_id(self, country_id) -> 'Country':
        VALID_POSITIVE_INT.validate(country_id, 'get_country_by_id', exc=ValueError)
        if country_id not in self._countries_by_id.keys():
            for country in self.countries:
                assert isinstance(country, Country)
                if country.country_id == country_id:
                    return country
            raise ValueError(country_id)
        else:
            return self._countries_by_id[country_id]

    @property
    def groups(self) -> typing.Generator['Group', None, None]:
        for country in self.countries:
            assert isinstance(country, Country)
            for group in country.groups:
                assert isinstance(group, Group)
                yield group

    @property
    def statics(self) -> typing.Generator['Static', None, None]:
        for country in self.countries:
            assert isinstance(country, Country)
            for static in country.statics:
                assert isinstance(static, Static)
                yield static

    @property
    def farps(self) -> typing.Generator['Static', None, None]:
        for static in self.statics:
            assert isinstance(static, Static)
            if static.static_is_farp:
                yield static

    def get_groups_from_category(self, category) -> typing.Generator['Group', None, None]:
        Mission.validator_group_category.validate(category, 'get_groups_from_category')
        for group in self.groups:
            assert isinstance(group, Group)
            if group.group_category == category:
                yield group

    @property
    def units(self) -> typing.Generator['BaseUnit', None, None]:
        for group in self.groups:
            assert isinstance(group, Group)
            for unit in group.units:
                yield unit

    def get_units_from_category(self, category) -> typing.Generator['BaseUnit', None, None]:
        Mission.validator_group_category.validate(category, 'group category')
        for unit in self.units:
            assert isinstance(unit, BaseUnit)
            if unit.group_category == category:
                yield unit

    def get_group_by_id(self, group_id) -> 'Group':
        VALID_POSITIVE_INT.validate(group_id, 'get_group_by_id')
        for group in self.groups:
            assert isinstance(group, Group)
            if group.group_id == group_id:
                return group

    def get_group_by_name(self, group_name) -> 'Group':
        VALID_STR.validate(group_name, 'get_group_by_name')
        for group in self.groups:
            assert isinstance(group, Group)
            if group.group_name == group_name:
                return group

    def get_unit_by_name(self, unit_name) -> 'BaseUnit':
        VALID_STR.validate(unit_name, 'get_unit_by_name')
        for unit in self.units:
            assert isinstance(unit, BaseUnit)
            if unit.unit_name == unit_name:
                return unit

    def get_unit_by_id(self, unit_id) -> typing.Union['BaseUnit', None]:
        VALID_POSITIVE_INT.validate(unit_id, 'get_unit_by_id')
        for unit in self.units:
            assert isinstance(unit, BaseUnit)
            if unit.unit_id == unit_id:
                return unit
        else:
            return None


class Trig(BaseMissionObject):
    def __init__(self, mission_dict, l10n):
        super().__init__(mission_dict, l10n)

    @property
    def _section_trig(self):
        return self.d['trig']


class Result(BaseMissionObject):
    def __init__(self, mission_dict, l10n):
        super().__init__(mission_dict, l10n)

    @property
    def _section_result(self):
        return self.d['result']


class GroundControl(BaseMissionObject):
    validator_commander = Validator(_type=int, _min=0, _max=100, exc=ValueError, logger=LOGGER)

    def __init__(self, mission_dict, l10n):
        super().__init__(mission_dict, l10n)

    def __repr__(self):
        return 'GroundControl({})'.format(self._section_ground_control)

    @property
    def _section_ground_control(self):
        return self.d['groundControl']

    @property
    def _section_ground_control_roles(self):
        return self.d['groundControl']['roles']

    @property
    def pilots_control_vehicles(self):
        return self._section_ground_control['isPilotControlVehicles']

    @pilots_control_vehicles.setter
    def pilots_control_vehicles(self, value):
        VALID_BOOL.validate(value, 'pilots_control_vehicles')
        self._section_ground_control['isPilotControlVehicles'] = value

    @property
    def _section_artillery_commander(self):
        return self._section_ground_control_roles['artillery_commander']

    @property
    def artillery_commander_red(self):
        return self._section_artillery_commander['red']

    @artillery_commander_red.setter
    def artillery_commander_red(self, value):
        self.validator_commander.validate(value, 'artillery_commander_red')
        self._section_artillery_commander['red'] = value

    @property
    def instructor_blue(self):
        return self._section_instructor['blue']

    @instructor_blue.setter
    def instructor_blue(self, value):
        self.validator_commander.validate(value, 'instructor_blue')
        self._section_instructor['blue'] = value

    @property
    def instructor_red(self):
        return self._section_instructor['red']

    @instructor_red.setter
    def instructor_red(self, value):
        self.validator_commander.validate(value, 'instructor_red')
        self._section_instructor['red'] = value

    @property
    def _section_observer(self):
        return self._section_ground_control_roles['observer']

    @property
    def observer_blue(self):
        return self._section_observer['blue']

    @observer_blue.setter
    def observer_blue(self, value):
        self.validator_commander.validate(value, 'observer_blue')
        self._section_observer['blue'] = value

    @property
    def observer_red(self):
        return self._section_observer['red']

    @observer_red.setter
    def observer_red(self, value):
        self.validator_commander.validate(value, 'observer_red')
        self._section_observer['red'] = value

    @property
    def _section_forward_observer(self):
        return self._section_ground_control_roles['forward_observer']

    @property
    def forward_observer_blue(self):
        return self._section_forward_observer['blue']

    @forward_observer_blue.setter
    def forward_observer_blue(self, value):
        self.validator_commander.validate(value, 'forward_observer_blue')
        self._section_forward_observer['blue'] = value

    @property
    def forward_observer_red(self):
        return self._section_forward_observer['red']

    @forward_observer_red.setter
    def forward_observer_red(self, value):
        self.validator_commander.validate(value, 'forward_observer_red')
        self._section_forward_observer['red'] = value

    @property
    def artillery_commander_blue(self):
        return self._section_artillery_commander['blue']

    @artillery_commander_blue.setter
    def artillery_commander_blue(self, value):
        self.validator_commander.validate(value, 'artillery_commander_blue')
        self._section_artillery_commander['blue'] = value

    @property
    def _section_instructor(self):
        return self._section_ground_control_roles['instructor']


# noinspection PyProtectedMember
class Weather(BaseMissionObject):
    validator_precipitations = Validator(_type=int, _min=0, _max=4, exc=ValueError, logger=LOGGER)
    validator_cloud_density = Validator(_type=int, _min=0, _max=10, exc=ValueError, logger=LOGGER)
    validator_cloud_thickness = Validator(_type=int, _min=200, _max=2000, exc=ValueError,
                                          logger=LOGGER)
    validator_cloud_base = Validator(_type=int, _min=300, _max=5000, exc=ValueError, logger=LOGGER)
    validator_fog_visibility = Validator(_type=int, _min=0, _max=6000, exc=ValueError,
                                         logger=LOGGER)
    validator_visibility = Validator(_type=int, _min=0, _max=80000, exc=ValueError,
                                     logger=LOGGER)
    validator_fog_thickness = Validator(_type=int, _min=0, _max=1000, exc=ValueError,
                                        logger=LOGGER)
    validator_qnh = Validator(_type=int, _min=720, _max=790, exc=ValueError, logger=LOGGER)
    validator_temperature = Validator(_type=int, _min=-50, _max=50, exc=ValueError, logger=LOGGER)
    validator_temp_spring_or_fall = Validator(_type=int, _min=-10, _max=30, exc=ValueError,
                                              logger=LOGGER)
    validator_temp_winter = Validator(_type=int, _min=-50, _max=15, exc=ValueError, logger=LOGGER)
    validator_temp_summer = Validator(_type=int, _min=5, _max=50, exc=ValueError, logger=LOGGER)
    seasons_enum = {
        1: {
            'name': 'summer',
            'temp_validator': validator_temp_summer,
        },
        2: {
            'name': 'winter',
            'temp_validator': validator_temp_winter,
        },
        3: {
            'name': 'spring',
            'temp_validator': validator_temp_spring_or_fall,
        },
        4: {
            'name': 'fall',
            'temp_validator': validator_temp_spring_or_fall,
        },
        'summer': 1,
        'winter': 2,
        'spring': 3,
        'fall': 4,
    }
    validator_season_name = Validator(_type=str, _in_list=['summer', 'winter', 'fall', 'spring'],
                                      exc=ValueError, logger=LOGGER)
    validator_season_code = Validator(_type=int, _min=1, _max=4, exc=ValueError, logger=LOGGER)
    validator_turbulence = Validator(_type=int, _min=0, _max=60, exc=ValueError, logger=LOGGER)
    validator_atmo_type = Validator(_type=int, _min=0, _max=1, exc=ValueError, logger=LOGGER)
    validator_wind_speed = Validator(_type=int, _min=0, _max=50, exc=ValueError, logger=LOGGER)

    def __init__(self, mission_dict, l10n):
        super().__init__(mission_dict, l10n)

    def __repr__(self):
        return 'Weather({})'.format(self._section_weather)

    def __eq__(self, other):
        if not isinstance(other, Weather):
            raise ValueError('"other" must be an Weather instance; got: {}'.format(type(other)))
        return self._section_weather == other._section_weather

    def get_season_code_from_name(self, season_name):
        self.validator_season_name.validate(season_name, 'get_season_code_from_name')
        return self.seasons_enum[season_name]

    @property
    def _section_wind_at_ground_level(self):
        return self._section_wind['atGround']

    @property
    def turbulence_at_ground_level(self):
        return self._section_weather['groundTurbulence']

    @turbulence_at_ground_level.setter
    def turbulence_at_ground_level(self, value):
        self.validator_turbulence.validate(value, 'turbulence_at_ground_level')
        self._section_weather['groundTurbulence'] = value

    @property
    def wind_at_ground_level_speed(self):
        return self._section_wind_at_ground_level['speed']

    @wind_at_ground_level_speed.setter
    def wind_at_ground_level_speed(self, value):
        self.validator_wind_speed.validate(value, 'wind_at_ground_level_speed')
        self._section_wind_at_ground_level['speed'] = value

    @property
    def _section_fog(self):
        return self._section_weather['fog']

    @property
    def fog_thickness(self):
        return self._section_fog['thickness']

    @fog_thickness.setter
    def fog_thickness(self, value):
        self.validator_fog_thickness.validate(value, 'fog_thickness')
        self._section_fog['thickness'] = value

    @property
    def fog_visibility(self):
        return self._section_fog['visibility']

    @fog_visibility.setter
    def fog_visibility(self, value):
        self.validator_fog_visibility.validate(value, 'fog_visibility')
        self._section_fog['visibility'] = value

    @property
    def fog_enabled(self):
        return self._section_weather['enable_fog']

    @fog_enabled.setter
    def fog_enabled(self, value):
        VALID_BOOL.validate(value, 'enable_fog')
        self._section_weather['enable_fog'] = value

    @property
    def _section_visibility(self):
        return self._section_weather['visibility']

    @property
    def visibility(self):
        return self._section_visibility['distance']

    @visibility.setter
    def visibility(self, value):
        self.validator_visibility.validate(value, 'visibility')
        self._section_visibility['distance'] = value

    @property
    def precipitations(self):
        return self._section_clouds['iprecptns']

    @precipitations.setter
    def precipitations(self, value):
        self.validator_precipitations.validate(value, 'precipitations')
        if value > 0 and self.cloud_density <= 4:
            raise ValueError('No rain or snow if cloud density is less than 5')
        if value in [2, 4] and self.cloud_density <= 8:
            raise ValueError('No thunderstorm or snowstorm if cloud density is less than 9')
        if value > 2 and self.temperature > 0:
            raise ValueError('No snow with temperature over 0; use rain or thunderstorm instead')
        if value in [1, 2] and self.temperature < 0:
            raise ValueError(
                'No rain or thunderstorm if temperature is below 0; use snow or snowstorm instead')
        self._section_clouds['iprecptns'] = value

    @property
    def wind_at8000_dir(self):
        return self._section_wind_at8000['speed']

    @wind_at8000_dir.setter
    def wind_at8000_dir(self, value):
        Mission.validator_heading.validate(value, 'wind_at8000_dir')
        self._section_wind_at8000['dir'] = value

    @property
    def _section_weather(self):
        return self.d['weather']

    @property
    def temperature(self):
        return self._section_season['temperature']

    @temperature.setter
    def temperature(self, value):
        self.validator_temperature.validate(value, 'temperature')
        self._section_season['temperature'] = value
        # if value > 0 and self.precipitations > 2:
        #     self.precipitations -= 2
        # if value < 0 < self.precipitations < 3:  # PyKek
        #     self.precipitations += 2

    @property
    def _section_wind_at8000(self):
        return self._section_wind['at8000']

    @property
    def wind_at_ground_level_dir(self):
        return self._section_wind_at_ground_level['dir']

    @wind_at_ground_level_dir.setter
    def wind_at_ground_level_dir(self, value):
        Mission.validator_heading.validate(value, 'wind_at_ground_level_dir')
        self._section_wind_at_ground_level['dir'] = value

    @property
    def _section_wind(self):
        return self._section_weather['wind']

    @property
    def _section_wind_at2000(self):
        return self._section_wind['at2000']

    @property
    def season_name(self):
        return self.seasons_enum[self.season_code]['name']

    @property
    def qnh(self):
        return self._section_weather['qnh']

    @qnh.setter
    def qnh(self, value):
        self.validator_qnh.validate(value, 'qnh')
        self._section_weather['qnh'] = value

    @property
    def wind_at2000_speed(self):
        return self._section_wind_at2000['speed']

    @wind_at2000_speed.setter
    def wind_at2000_speed(self, value):
        self.validator_wind_speed.validate(value, 'wind_at2000_speed')
        self._section_wind_at2000['speed'] = value

    @property
    def wind_at2000_dir(self):
        return self._section_wind_at2000['dir']

    @wind_at2000_dir.setter
    def wind_at2000_dir(self, value):
        Mission.validator_heading.validate(value, 'wind_at2000_dir')
        self._section_wind_at2000['dir'] = value

    @property
    def _section_season(self):
        return self._section_weather['season']

    @property
    def atmosphere_type(self):
        return self._section_weather['atmosphere_type']

    @atmosphere_type.setter
    def atmosphere_type(self, value):
        self.validator_atmo_type.validate(value, 'atmosphere_type')
        self._section_weather['atmosphere_type'] = value

    @property
    def season_code(self):
        return self._section_season['iseason']

    @season_code.setter
    def season_code(self, value):
        self.validator_season_code.validate(value, 'season')
        self._section_season['iseason'] = value
        if self.temperature < self.seasons_enum[value]['temp_validator'].min:
            self.temperature = self.seasons_enum[value]['temp_validator'].min
        if self.temperature > self.seasons_enum[value]['temp_validator'].max:
            self.temperature = self.seasons_enum[value]['temp_validator'].max

    @property
    def _section_clouds(self):
        return self._section_weather['clouds']

    @property
    def cloud_thickness(self):
        return self._section_clouds['thickness']

    @cloud_thickness.setter
    def cloud_thickness(self, value):
        self.validator_cloud_thickness.validate(value, 'cloud_thickness')
        self._section_clouds['thickness'] = value

    @property
    def cloud_base(self):
        return self._section_clouds['base']

    @cloud_base.setter
    def cloud_base(self, value):
        self.validator_cloud_base.validate(value, 'cloud_base')
        self._section_clouds['base'] = value

    @property
    def cloud_density(self):
        return self._section_clouds['density']

    @cloud_density.setter
    def cloud_density(self, value):
        self.validator_cloud_density.validate(value, 'cloud_density')
        self._section_clouds['density'] = value

    @property
    def wind_at8000_speed(self):
        return self._section_wind_at8000['speed']

    @wind_at8000_speed.setter
    def wind_at8000_speed(self, value):
        self.validator_wind_speed.validate(value, 'wind_at8000_speed')
        self._section_wind_at8000['speed'] = value


# noinspection PyProtectedMember
class Country(Coalition):
    def __init__(self, mission_dict, l10n, coa_color, country_index):
        super().__init__(mission_dict, l10n, coa_color)
        self.__groups = {
            'helicopter': {},
            'plane': {},
            'vehicle': {},
            'ship': {},
        }
        self.country_index = country_index
        self.__static = {}

    def __repr__(self):
        return 'Country({}, {}, {})'.format(self._section_country, self.coa_color, self.country_index)

    def __eq__(self, other):
        if not isinstance(other, Country):
            raise ValueError('"other" must be an Country instance; got: {}'.format(type(other)))
        return self._section_country == other._section_country

    @property
    def _section_this_country(self):
        return self._section_coalition['country'][self.country_index]

    @property
    def country_id(self):
        return self._section_this_country['id']

    @property
    def country_name(self):
        return self._section_this_country['name']

    @property
    def groups(self) -> typing.Generator['Group', None, None]:
        for group_category in Mission.valid_group_categories:
            if group_category in self._section_this_country.keys():
                for group_index in self._section_this_country[group_category]['group']:
                    if group_index not in self.__groups[group_category]:
                        self.__groups[group_category][group_index] = Group(self.d, self.l10n, self.coa_color,
                                                                           self.country_index, group_category,
                                                                           group_index)
                    yield self.__groups[group_category][group_index]

    @property
    def statics(self) -> typing.Generator['Static', None, None]:
        if 'static' in self._section_this_country.keys():
            for static_index in self._section_this_country['static']['group']:
                if static_index not in self.__static:
                    self.__static[static_index] = Static(self.d, self.l10n, self.coa_color,
                                                         self.country_index, static_index)
                yield self.__static[static_index]

    def get_groups_from_category(self, category) -> typing.Generator['Group', None, None]:
        Mission.validator_group_category.validate(category, 'get_groups_from_category')
        for group in self.groups:
            assert isinstance(group, Group)
            if group.group_category == category:
                yield group

    def get_group_by_id(self, group_id) -> 'Group':
        for group in self.groups:
            assert isinstance(group, Group)
            if group.group_id == group_id:
                return group

    def get_group_by_name(self, group_name) -> 'Group':
        for group in self.groups:
            assert isinstance(group, Group)
            if group.group_name == group_name:
                return group

    @property
    def units(self) -> typing.Generator['BaseUnit', None, None]:
        for group in self.groups:
            assert isinstance(group, Group)
            for unit in group.units:
                yield unit

    def get_unit_by_name(self, unit_name) -> 'BaseUnit':
        for unit in self.units:
            assert isinstance(unit, BaseUnit)
            if unit.unit_name == unit_name:
                return unit

    def get_unit_by_id(self, unit_id) -> 'BaseUnit':
        for unit in self.units:
            assert isinstance(unit, BaseUnit)
            if unit.unit_id == unit_id:
                return unit

    def get_units_from_category(self, category) -> typing.Generator['BaseUnit', None, None]:
        Mission.validator_group_category.validate(category, 'group category')
        for unit in self.units:
            assert isinstance(unit, BaseUnit)
            if unit.group_category == category:
                yield unit


class Static(Country):
    def __init__(self, mission_dict, l10n, coa_color, country_index, static_index):
        super(Static, self).__init__(mission_dict, l10n, coa_color, country_index)
        self.static_index = static_index

    @property
    def static_id(self):
        return self._section_static['groupId']

    @static_id.setter
    def static_id(self, value):
        VALID_INT.validate(value, 'groupId')
        self._section_static['groupId'] = value

    @property
    def _section_static(self):
        return self._section_this_country['static']['group'][self.static_index]

    @property
    def _static_name_key(self):
        return self._section_static['name']

    @property
    def static_name(self):
        return self.l10n[self._static_name_key]

    @static_name.setter
    def static_name(self, value):
        validator_group_or_unit_name.validate(value, 'group name')
        self.l10n[self._static_name_key] = value

    @property
    def static_category(self):
        return self._section_static['units'][1]['category']

    @property
    def static_is_farp(self):
        return self.static_category == 'Heliports'

    @property
    def static_position(self):
        unit = self._section_static['units'][1]
        return unit['x'], unit['y']


# noinspection PyProtectedMember
class Group(Country):
    attribs = ('group_category', 'group_index', 'group_hidden', 'group_start_time', '_group_name_key')

    class Route:

        class Point:
            def __init__(self, parent_route):
                assert isinstance(parent_route, Group.Route)
                self.parent_route = parent_route

            def __repr__(self):
                return 'Route({})'.format(self.parent_route.parent_group.group_name)

        def __init__(self, parent_group):
            assert isinstance(parent_group, Group)
            self.parent_group = parent_group

        def __repr__(self):
            return 'Route({})'.format(self.parent_group.group_name)

        @property
        def _section_route(self):
            return self.parent_group._section_group['route']['points']

        @property
        def points(self):
            raise NotImplementedError('uh')

    validator_group_route = Validator(_type=Route, exc=ValueError, logger=LOGGER)
    units_class_enum = None

    def __init__(self, mission_dict, l10n, coa_color, country_index, group_category, group_index):
        super().__init__(mission_dict, l10n, coa_color, country_index)
        self.group_category = group_category
        self.group_index = group_index
        self.__group_route = None
        self.__units = {}
        self.units_class_enum = {
            'helicopter': Helicopter,
            'plane': Plane,
            'ship': Ship,
            'vehicle': Vehicle,
        }

    def __repr__(self):
        return 'Group({}, {}, {}, {}, {})'.format(self._section_group, self.coa_color, self.country_index,
                                                  self.group_category, self.group_index)

    def __eq__(self, other):
        if not isinstance(other, Group):
            raise ValueError(
                '"other" must be an AbstractUnit instance; got: {}'.format(type(other)))
        return self._section_group == other._section_group

    # noinspection PyTypeChecker
    @property
    def group_route(self) -> 'Group.Route':
        #  TODO
        if self.__group_route is None:
            self.__group_route = Group.Route(self)
        return self.__group_route

    @group_route.setter
    def group_route(self, value):
        self.validator_group_route.validate(value, 'group_route')
        self.__group_route = value

    @property
    def _section_group(self):
        return self._section_this_country[self.group_category]['group'][self.group_index]

    @property
    def _group_name_key(self):
        return self._section_group['name']

    @property
    def group_name(self):
        return self.l10n[self._group_name_key]

    @group_name.setter
    def group_name(self, value):
        validator_group_or_unit_name.validate(value, 'group name')
        self.l10n[self._group_name_key] = value

    @property
    def group_hidden(self):
        return self._section_group['hidden']

    @group_hidden.setter
    def group_hidden(self, value):
        VALID_BOOL.validate(value, 'property "hidden" for group')
        self._section_group['hidden'] = value

    @property
    def group_id(self):
        return self._section_group['groupId']

    @group_id.setter
    def group_id(self, value):
        VALID_INT.validate(value, 'groupId')
        self._section_group['groupId'] = value

    @property
    def group_start_delay(self):
        return self._section_group['start_time']

    @group_start_delay.setter
    def group_start_delay(self, value):
        VALID_INT.validate(value, 'group_start_delay')
        if value < 0:
            raise ValueError(self.group_name)
        self._section_group['start_time'] = value

    @property
    def group_start_time(self):
        return self.group_start_delay + self.mission_start_time

    @group_start_time.setter
    def group_start_time(self, value):
        VALID_INT.validate(value, 'group_start_time')
        self.group_start_delay = value - self.mission_start_time

    @property
    def group_start_date_time_as_string(self):
        return self._start_datetime_as_string(self.group_start_time)

    @property
    def units(self):
        for unit_index in self._section_group['units']:
            if unit_index not in self.__units.keys():
                self.__units[unit_index] = self.units_class_enum[self.group_category](self.d, self.l10n, self.coa_color,
                                                                                      self.country_index,
                                                                                      self.group_category,
                                                                                      self.group_index, unit_index)
            yield self.__units[unit_index]

    @property
    def first_unit(self) -> 'BaseUnit':
        return list(self.units)[0]

    def group_size(self) -> int:
        return len(list(self.units))

    def get_unit_by_name(self, unit_name):
        for unit in self.units:
            assert isinstance(unit, BaseUnit)
            if unit.unit_name == unit_name:
                return unit
        return None

    def get_unit_by_id(self, unit_id):
        for unit in self.units:
            assert isinstance(unit, BaseUnit)
            if unit.unit_id == unit_id:
                return unit
        return None

    def get_unit_by_index(self, unit_index):
        if unit_index in self._section_group['units'].keys():
            if unit_index not in self.__units.keys():
                self.__units[unit_index] = self.units_class_enum[self.group_category](self.d, self.l10n, self.coa_color,
                                                                                      self.country_index,
                                                                                      self.group_category,
                                                                                      self.group_index, unit_index)
            return self.__units[unit_index]
        return None

    @property
    def group_is_client_group(self):
        # TODO create test
        first_unit = self.get_unit_by_index(1)
        assert isinstance(first_unit, BaseUnit)
        return first_unit.skill == 'Client'

    @property
    def group_start_position(self):
        return self.group_route._section_route[1]['action']


# noinspection PyProtectedMember
class BaseUnit(Group):
    validator_skill = Validator(_type=str,
                                _in_list=['Average', 'Good', 'High', 'Excellent', 'Random', 'Client', 'Player'],
                                exc=ValueError, logger=LOGGER)
    validator_unit_types = Validator(_type=str, _in_list=[], exc=ValueError, logger=LOGGER)

    def __init__(self, mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index):
        super().__init__(mission_dict, l10n, coa_color, country_index, group_category, group_index)
        self.unit_index = unit_index

    def __repr__(self):
        return '{}({}, {}, {}, {}, {}, {})'.format(self.__class__.__name__, self._section_unit, self.coa_color,
                                                   self.country_index, self.group_category, self.group_index,
                                                   self.unit_index)

    @property
    def _section_unit(self):
        return self._section_group['units'][self.unit_index]

    @property
    def _unit_name_key(self):
        return self._section_unit['name']

    @property
    def unit_name(self):
        return self.l10n[self._unit_name_key]

    @unit_name.setter
    def unit_name(self, value):
        validator_group_or_unit_name.validate(value, 'unit name')
        self.l10n[self._unit_name_key] = value

    @property
    def skill(self):
        return self._section_unit['skill']

    @skill.setter
    def skill(self, value):
        self.validator_skill.validate(value, 'unit skill')
        self._section_unit['skill'] = value

    @property
    def speed(self):
        return self._section_unit['speed']

    @speed.setter
    def speed(self, value):
        VALID_FLOAT.validate(value, 'unit speed')
        self._section_unit['speed'] = value

    @property
    def unit_type(self):
        return self._section_unit['type']

    @unit_type.setter
    def unit_type(self, value):
        self.validator_unit_types.validate(value, 'unit type')
        self._section_unit['type'] = value

    @property
    def unit_id(self):
        return self._section_unit['unitId']

    @unit_id.setter
    def unit_id(self, value):
        VALID_INT.validate(value, 'unitId')
        self._section_unit['unitId'] = value

    @property
    def unit_pos_x(self):
        return float(self._section_unit['x'])

    @unit_pos_x.setter
    def unit_pos_x(self, value):
        VALID_FLOAT.validate(value, 'unit position X coordinate')
        self._section_unit['x'] = value

    @property
    def unit_pos_y(self):
        return float(self._section_unit['y'])

    @unit_pos_y.setter
    def unit_pos_y(self, value):
        VALID_FLOAT.validate(value, 'unit position Y coordinate')
        self._section_unit['y'] = value

    @property
    def unit_position(self):
        return self.unit_pos_x, self.unit_pos_y

    @unit_position.setter
    def unit_position(self, value):
        self.unit_pos_x, self.unit_pos_y = value

    @property
    def heading(self):
        return self._section_unit['heading']

    @heading.setter
    def heading(self, value):
        Mission.validator_heading.validate(value, 'unit heading')
        self._section_unit['heading'] = value

    @property
    def radio_presets(self) -> typing.Generator['FlyingUnit.RadioPresets', None, None]:
        raise TypeError('unit #{}: {}'.format(self.unit_id, self.unit_name))

    @property
    def has_radio_presets(self):
        return all([self.skill == 'Client', self.unit_type in FlyingUnit.RadioPresets.radio_enum.keys()])

    def __eq__(self, other):
        if not isinstance(other, BaseUnit):
            raise ValueError(
                '"other" must be an AbstractUnit instance; got: {}'.format(type(other)))
        return self._section_unit == other._section_unit


class FlyingUnit(BaseUnit):
    validator_board_number = Validator(_type=str, _regex=r'[0-9]{3}', exc=ValueError,
                                       logger=LOGGER)

    class RadioPresets:

        radio_enum = {
            'Ka-50': {
                1: {
                    'radio_name': 'R828',
                    'min': 20,
                    'max': 59.9,
                    'channels_qty': 10,
                },
                2: {
                    'radio_name': 'ARK22',
                    'min': 0.15,
                    'max': 1.75,
                    'channels_qty': 16,
                },
            },
            'Mi-8MT': {
                1: {
                    'radio_name': 'R863',
                    'min': 100,
                    'max': 399.9,
                    'channels_qty': 20,
                },
                2: {
                    'radio_name': 'R828',
                    'min': 20,
                    'max': 59.9,
                    'channels_qty': 10,
                },
            },
            'UH-1H': {
                1: {
                    'radio_name': 'ARC51',
                    'min': 225,
                    'max': 399.97,
                    'channels_qty': 20,
                },
            },
            'F-86F Sabre': {
                1: {
                    'radio_name': 'ARC-27',
                    'min': 225,
                    'max': 399.9,
                    'channels_qty': 18,
                },
            },
            'M-2000C': {
                1: {
                    'radio_name': 'UHF',
                    'min': 225,
                    'max': 400,
                    'channels_qty': 20,
                },
                2: {
                    'radio_name': 'V/UHF',
                    'min': 118,
                    'max': 400,
                    'channels_qty': 20,
                },
            },
            'MiG-21Bis': {
                1: {
                    'radio_name': 'R-832',
                    'min': 80,
                    'max': 399.9,
                    'channels_qty': 20,
                },
            },
            'P-51D': {
                1: {
                    'radio_name': 'SCR552',
                    'min': 100,
                    'max': 156,
                    'channels_qty': 4,
                },
            },
            'TF-51D': {
                1: {
                    'radio_name': 'SCR552',
                    'min': 100,
                    'max': 156,
                    'channels_qty': 4,
                },
            },
            'SpitfireLFMkIX': {
                1: {
                    'radio_name': 'SCR522',
                    'min': 100,
                    'max': 156,
                    'channels_qty': 4,
                },
            },
            'Bf-109K-4': {
                1: {
                    'radio_name': 'FuG 16 ZY',
                    'min': 38,
                    'max': 156,
                    'channels_qty': 5,
                },
            },
            'FW-190D9': {
                1: {
                    'radio_name': 'FuG 16',
                    'min': 38.4,
                    'max': 42.4,
                    'channels_qty': 4,
                },
            },
            'SA342L': {
                1: {
                    'radio_name': 'FM Radio',
                    'min': 30,
                    'max': 87.975,
                    'channels_qty': 8,
                },
            },
            'SA342M': {
                1: {
                    'radio_name': 'FM Radio',
                    'min': 30,
                    'max': 87.975,
                    'channels_qty': 8,
                },
            },
            'SA342Mistral': {
                1: {
                    'radio_name': 'FM Radio',
                    'min': 30,
                    'max': 87.975,
                    'channels_qty': 8,
                },
            },
        }

        def __init__(self, parent_unit, radio_num):
            assert isinstance(parent_unit, FlyingUnit)
            self.parent_unit = parent_unit
            self.radio_num = radio_num

        def __eq__(self, other):
            if not isinstance(other, FlyingUnit.RadioPresets):
                raise Exception('cannot compare RadioPreset instance with other object of type {}'.format(type(other)))
            assert isinstance(other, FlyingUnit.RadioPresets)
            if not self.radio_name == other.radio_name:
                return False
            for channel, frequency in self.channels:
                if not frequency == other.get_frequency(channel):
                    return False
            return True

        @property
        def radio_name(self):
            return self.radio_enum[self.parent_unit.unit_type][self.radio_num]['radio_name']

        @property
        def channels_qty(self):
            return self.radio_enum[self.parent_unit.unit_type][self.radio_num]['channels_qty']

        @property
        def min(self):
            return float(self.radio_enum[self.parent_unit.unit_type][self.radio_num]['min'])

        @property
        def max(self):
            return float(self.radio_enum[self.parent_unit.unit_type][self.radio_num]['max'])

        @property
        def _section_radio(self):
            return self.parent_unit._section_unit['Radio']

        @property
        def _section_channels(self):
            return self._section_radio[self.radio_num]['channels']

        @property
        def channels(self) -> typing.Generator[tuple, None, None]:
            for k in self._section_channels:
                yield (k, float(self._section_channels[k]))

        def get_frequency(self, channel):
            VALID_POSITIVE_INT.validate(channel, 'get_frequency')
            if 1 <= channel <= self.channels_qty:
                return float(self._section_channels[channel])
            else:
                raise ValueError(
                    'channel {} for radio {} in aircraft {}'.format(channel, self.radio_name,
                                                                    self.parent_unit.unit_name))

        def set_frequency(self, channel, frequency):
            VALID_POSITIVE_INT.validate(channel, 'set_frequency')
            VALID_FLOAT.validate(frequency, 'set_frequency')
            if 1 <= channel <= self.channels_qty:
                # noinspection PyTypeChecker
                if self.min <= frequency <= self.max:
                    self._section_channels[channel] = float(frequency)
                else:
                    raise ValueError(
                        'frequency {} for channel {} for radio {} in aircraft {}'.format(frequency, channel,
                                                                                         self.radio_name,
                                                                                         self.parent_unit.unit_name))
            else:
                raise ValueError(
                    'channel {} for radio {} in aircraft {}'.format(channel, self.radio_name,
                                                                    self.parent_unit.unit_name))

    def __init__(self, mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index):
        super().__init__(mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index)

    @property
    def radio_presets(self) -> typing.Generator['FlyingUnit.RadioPresets', None, None]:
        if self.skill == 'Client' and self.unit_type in FlyingUnit.RadioPresets.radio_enum.keys():
            for k in self._section_unit['Radio']:
                yield FlyingUnit.RadioPresets(self, k)
        else:
            raise TypeError('unit #{}: {}'.format(self.unit_id, self.unit_name))

    @property
    def radios(self):
        try:
            return self._section_unit['Radio']
        except KeyError:
            raise KeyError(self.unit_type)

    def get_radio_by_name(self, radio_name):
        if self.has_radio_presets:
            for k in FlyingUnit.RadioPresets.radio_enum[self.unit_type].keys():
                if radio_name == FlyingUnit.RadioPresets.radio_enum[self.unit_type][k]['radio_name']:
                    return FlyingUnit.RadioPresets(self, k)
            raise TypeError('{} for aircraft: {}'.format(radio_name, self.unit_type))
        else:
            raise TypeError('unit #{}: {}'.format(self.unit_id, self.unit_name))

    def get_radio_by_number(self, radio_number):
        if self.has_radio_presets:
            if radio_number in FlyingUnit.RadioPresets.radio_enum[self.unit_type].keys():
                return FlyingUnit.RadioPresets(self, radio_number)
            else:
                raise TypeError(
                    'radio number {} for aircraft: {}'.format(radio_number, self.unit_type))
        else:
            raise TypeError('unit #{}: {}'.format(self.unit_id, self.unit_name))

    @property
    def livery(self):
        return self._section_unit['livery_id']

    @livery.setter
    def livery(self, value):
        # TODO validate livery_id
        VALID_STR.validate(value, 'unit livery')
        self._section_unit['livery_id'] = value

    @property
    def onboard_num(self):
        return self._section_unit['onboard_num']

    @onboard_num.setter
    def onboard_num(self, value):
        FlyingUnit.validator_board_number.validate(value, 'unit onboard number')
        self._section_unit['onboard_num'] = value


class Helicopter(FlyingUnit):
    def __init__(self, mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index):
        super().__init__(mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index)


class Plane(FlyingUnit):
    def __init__(self, mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index):
        super().__init__(mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index)


class Vehicle(BaseUnit):
    def __init__(self, mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index):
        super().__init__(mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index)


class Ship(BaseUnit):
    def __init__(self, mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index):
        super().__init__(mission_dict, l10n, coa_color, country_index, group_category, group_index, unit_index)



# coding=utf-8

import math
import pickle
from collections import namedtuple

# pylint: skip-file
# noinspection PyProtectedMember
from emiz._parking_spots import parkings
from emiz.mission import Static

ParkingSpot = namedtuple('ParkingSpot', 'airport spot')
parkings = pickle.loads(parkings)


def clear_farps():
    parkings['FARP'] = {}


def add_farp(farp: Static):
    parkings['FARP'][farp.static_name] = farp.static_position


def unit_pos_to_spot(unit_pos) -> ParkingSpot:
    min_ = 50
    res = None
    for airport in parkings:
        for spot in parkings[airport]:
            spot_pos = parkings[airport][spot]
            dist = math.hypot(unit_pos[0] - spot_pos[0], unit_pos[1] - spot_pos[1])
            if dist < min_:
                min_ = dist
                res = ParkingSpot(airport=airport, spot=spot)
    return res



# coding=utf-8
"""
Applies a METAR to a MIZ file
"""
import typing
from pathlib import Path

import elib.path
from metar import Metar

from emiz import MAIN_LOGGER
from emiz.miz import Miz

from .. import custom_metar, mission_weather

LOGGER = MAIN_LOGGER.getChild(__name__)


def set_weather_from_metar(
        metar: typing.Union[Metar.Metar, str],
        in_file: typing.Union[str, Path],
        out_file: typing.Union[str, Path] = None
) -> typing.Tuple[typing.Union[str, None], typing.Union[str, None]]:
    """
    Applies the weather from a METAR object to a MIZ file

    Args:
        metar: metar object
        in_file: path to MIZ file
        out_file: path to output MIZ file (will default to in_file)

    Returns: tuple of error, success

    """
    error, metar = custom_metar.CustomMetar.get_metar(metar)

    if error:
        return error, None

    LOGGER.debug(f'METAR: {metar.code}')

    in_file = elib.path.ensure_file(in_file)

    if out_file is None:
        out_file = in_file
    else:
        out_file = elib.path.ensure_file(out_file, must_exist=False)
    LOGGER.debug(f'applying metar: {in_file} -> {out_file}')

    try:
        LOGGER.debug('building MissionWeather')
        _mission_weather = mission_weather.MissionWeather(metar)

        with Miz(str(in_file)) as miz:
            _mission_weather.apply_to_miz(miz)
            miz.zip(str(out_file))
            return None, f'successfully applied METAR to {in_file}'

    except ValueError:
        error = f'Unable to apply METAR string to the mission.\n' \
                f'This is most likely due to a freak value, this feature is still experimental.\n' \
                f'I will fix it ASAP !'
        return error, None



# coding=utf-8
# pylint: disable=wrong-import-position
"""
Etcher's MIZ library
"""
from elib.custom_logging import get_logger
from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution('emiz').version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    __version__ = 'not installed'

MAIN_LOGGER = get_logger('EMIZ')
MAIN_LOGGER.info(f'EMIZ version {__version__}')

from . import weather, edit_miz
from .miz import Mission, Miz

__all__ = ['Miz', 'Mission', 'edit_miz', 'weather']



# coding=utf-8
"""
Manages date and time in a mission
"""
import datetime
import re

from emiz import MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)

RE_INPUT_STRING = re.compile(r'^'
                             r'(?P<year>[\d]{4})'
                             r'(?P<month>[\d]{2})'
                             r'(?P<day>[\d]{2})'
                             r'(?P<hour>[\d]{2})'
                             r'(?P<minute>[\d]{2})'
                             r'(?P<second>[\d]{2})'
                             r'$')


class MissionTime:
    """
    Represents date and time in a mission
    """

    def __init__(self, moment: datetime.datetime):
        self.date = datetime.date(moment.year, moment.month, moment.day)
        self.time = moment.hour * 3600 + moment.minute * 60 + moment.second

    def apply_to_miz(self, miz):
        """
        Applies this datetime to a Miz object (it will be mutated in place)

        Args:
            miz: MIZ object to mutate

        Returns: True

        """
        miz.mission.day = self.date.day
        miz.mission.month = self.date.month
        miz.mission.year = self.date.year
        miz.mission.mission_start_time = self.time

        return True

    @staticmethod
    def from_string(input_str) -> 'MissionTime':
        # noinspection SpellCheckingInspection
        """
        Creates a MissionTime instance from a string

        Format: YYYYMMDDHHMMSS

        Args:
            input_str: string to parse

        Returns: MissionTime instance

        """
        match = RE_INPUT_STRING.match(input_str)
        if not match:
            raise ValueError(f'badly formatted date/time: {input_str}')

        return MissionTime(
            datetime.datetime(
                int(match.group('year')),
                int(match.group('month')),
                int(match.group('day')),
                int(match.group('hour')),
                int(match.group('minute')),
                int(match.group('second')),
            )
        )

    @staticmethod
    def now() -> 'MissionTime':
        """

        Returns: MissionTime object with the current time

        """
        return MissionTime(datetime.datetime.now())



# coding=utf-8
"""
Represents the result of a query to AVWX API
"""
from collections import defaultdict

from emiz import MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)


# pylint: disable=too-few-public-methods
class _AVWXProp:
    """
    Represents a property of AVWX result (simple descriptor)
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, _):
        if obj is None:
            return self
        return obj.data[self.func.__name__]


# noinspection SpellCheckingInspection
# pylint: disable=too-many-public-methods
class AVWXResult:
    """
    Represents the result of a query to AVWX API
    """
    default_value = 'NOTSET'

    @staticmethod
    def default_factory():
        """Returns default value for unset property in the data dict"""
        return AVWXResult.default_value

    def __init__(self, **kwargs):
        self.data = defaultdict(default_factory=self.default_factory)
        self.data.update(kwargs)

    @_AVWXProp
    def altimeter(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def cloudlist(self) -> list:
        """AVWX result property"""
        pass

    @_AVWXProp
    def dewpoint(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def flightrules(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def info(self) -> dict:
        """AVWX result property"""
        pass

    @_AVWXProp
    def meta(self) -> dict:
        """AVWX result property"""
        pass

    @_AVWXProp
    def otherlist(self) -> list:
        """AVWX result property"""
        pass

    @_AVWXProp
    def rawreport(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def remarks(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def remarksinfo(self) -> dict:
        """AVWX result property"""
        pass

    @_AVWXProp
    def runwayvislist(self) -> list:
        """AVWX result property"""
        pass

    @_AVWXProp
    def speech(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def station(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def summary(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def temperature(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def time(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def translations(self) -> dict:
        """AVWX result property"""
        pass

    @_AVWXProp
    def units(self) -> dict:
        """AVWX result property"""
        pass

    @_AVWXProp
    def visibility(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def winddirection(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def windgust(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def windspeed(self) -> str:
        """AVWX result property"""
        pass

    @_AVWXProp
    def windvariabledir(self) -> list:
        """AVWX result property"""
        pass



# coding=utf-8

import tempfile
from datetime import datetime

import elib.path
import pytest
from mockito import mock, when

import emiz.miz
from emiz.miz import Mission
from emiz.weather.mizfile.mizfile_get_metar import get_metar_from_mission


@pytest.fixture(name='setup')
def _setup():
    miz = mock()
    mission = mock(spec=Mission)
    mission.weather = mock()
    mission.weather.wind_at_ground_level_dir = 180
    mission.weather.wind_at_ground_level_speed = 0
    mission.weather.visibility = 9999
    mission.weather.fog_visibility = 0
    mission.weather.fog_enabled = False
    mission.weather.precipitations = 0
    mission.weather.cloud_density = 0
    mission.weather.cloud_base = 8000
    mission.weather.temperature = 10
    mission.weather.qnh = 756
    miz.mission = mission
    time = 'time'
    icao = 'UGTB'
    when(emiz.miz.Miz).__enter__(...).thenReturn(miz)
    when(elib.path).ensure_file(...).thenReturn(elib.path.Path('test.miz'))
    when(emiz.miz.Miz).__exit__(...)
    when(tempfile).mkdtemp(...).thenReturn('no_tmp_dir')
    yield miz, mission, time, icao


# http://localhost:63342/EMIZ/htmlcov/emiz_weather_mizfile_mizfile_get_metar_py.html
def test_basic_metar(setup):
    _, _, time, icao = setup
    metar = get_metar_from_mission('test', time=time, icao=icao)
    assert metar == f'{icao} {time} 00000MPS CAVOK 10/10 Q1008 NOSIG'


def test_cavok(setup):
    _, mission, time, icao = setup
    mission.weather.visibility = 9998
    metar = get_metar_from_mission('test', time=time, icao=icao)
    assert 'CAVOK' not in metar


@pytest.mark.parametrize('density', range(0, 10))
def test_cloud_density(setup, density):
    _, mission, time, icao = setup
    mission.weather.cloud_density = density
    metar = get_metar_from_mission('test', time=time, icao=icao)
    if density == 0:
        assert metar == f'{icao} {time} 00000MPS CAVOK 10/10 Q1008 NOSIG'
    elif density < 4:
        assert metar == f'{icao} {time} 00000MPS CAVOK FEW262 10/10 Q1008 NOSIG'
    elif density < 7:
        assert metar == f'{icao} {time} 00000MPS CAVOK SCT262 10/10 Q1008 NOSIG'
    elif density < 9:
        assert metar == f'{icao} {time} 00000MPS CAVOK BKN262 10/10 Q1008 NOSIG'
    else:
        assert metar == f'{icao} {time} 00000MPS CAVOK OVC262 10/10 Q1008 NOSIG'


def test_no_time(setup):
    _, _, _, icao = setup
    now = datetime.utcnow()
    day = now.day
    hour = now.hour
    minute = now.minute
    time = f'{day:02}{hour:02}{minute:02}Z'
    metar = get_metar_from_mission('test', icao=icao)
    assert metar == f'{icao} {time} 00000MPS CAVOK 10/10 Q1008 NOSIG'


@pytest.mark.parametrize('fog_visibility', range(0, 6001, 500))
def test_fog(setup, fog_visibility):
    _, mission, time, icao = setup
    mission.weather.fog_visibility = fog_visibility
    mission.weather.fog_enabled = True
    metar = get_metar_from_mission('test', time=time, icao=icao)
    if fog_visibility < 5000:
        assert metar == f'{icao} {time} 00000MPS {fog_visibility:04d}M 10/10 Q1008 NOSIG'
    else:
        assert metar == f'{icao} {time} 00000MPS {fog_visibility:04d}M 10/10 Q1008 NOSIG'



# coding=utf-8
"""
Tests start time
"""

import pytest
from hypothesis import strategies as st
from hypothesis import given, settings

from emiz.miz import Miz


@given(time=st.integers(0, 86399))
@settings(max_examples=1)
def test_mission_start_time(time, mission):
    assert mission.mission_start_time == 43200
    assert mission.mission_start_datetime_as_string == '01/06/2011 12:00:00'
    mission.mission_start_time = 0
    assert mission.mission_start_datetime_as_string == '01/06/2011 00:00:00'
    mission.mission_start_time = time


@given(time=st.integers(min_value=86400))
@settings(max_examples=1)
def test_wrong_mission_start_time(time, mission):
    with pytest.raises(ValueError):
        mission.mission_start_time = time


@given(time=st.integers(max_value=-1))
@settings(max_examples=1)
def test_wrong_mission_start_time_neg(time, mission):
    with pytest.raises(ValueError):
        mission.mission_start_time = time


def test_group_start_time(test_file, out_file):
    with Miz(test_file) as miz:
        group = miz.mission.get_group_by_name('etcher')
        assert miz.mission.mission_start_time == group.group_start_time
        assert miz.mission.mission_start_datetime_as_string == group.group_start_date_time_as_string
        assert group.group_start_date_time_as_string == '01/06/2011 12:00:00'
        assert group.group_start_delay == 0
        group.group_start_delay += 60
        assert group.group_start_date_time_as_string == '01/06/2011 12:01:00'
        group.group_start_delay += 3600
        assert group.group_start_date_time_as_string == '01/06/2011 13:01:00'
        group.group_start_delay = 0
        group.group_start_delay = 3600
        assert group.group_start_date_time_as_string == '01/06/2011 13:00:00'
        miz.zip(out_file)

    with Miz(out_file) as miz:
        group = miz.mission.get_group_by_name('etcher')
        assert group.group_start_delay == 3600



# coding=utf-8

import os

import datadiff
import pytest

from emiz.sltp import SLTP, SLTPParsingError

if os.path.exists('./test_files'):
    BASE_PATH = os.path.abspath('./test_files/sltp')
elif os.path.exists('./test/test_files'):
    BASE_PATH = os.path.abspath('./test/test_files/sltp')
else:
    raise RuntimeError('cannot find test files')

TEST_FILES_DIR = os.path.join(BASE_PATH, 'pass')
TEST_FILES_DIR_FAIL = os.path.join(BASE_PATH, 'fail')
TEST_FILES_DIR_DIFF = os.path.join(BASE_PATH, 'diff')
TEST_FILES_DIR_LONG = os.path.join(BASE_PATH, 'long')

ENCODING = 'iso8859_15'


def _assert_same(input_, output):

    for x in input_:
        try:
            assert x in output
        except AssertionError:
            print(datadiff.diff(input_, output))
            raise

    assert len(input_) == len(output)


def _assert_different(input_, output):

    for x in input_:
        try:
            assert x in output
        except AssertionError:
            break
    else:
        pytest.fail('resulting dicts should be different')


def _do_test(test_file, compare_func):
    parser = SLTP()
    with open(test_file, encoding=ENCODING) as f:
        data = f.read()
    decoded_data, qualifier = parser.decode(data)

    parser = SLTP()
    encoded_data = parser.encode(decoded_data, qualifier)

    output = encoded_data.split('\n')
    input_ = data.split('\n')

    compare_func(input_, output)


@pytest.mark.parametrize('test_file', os.listdir(TEST_FILES_DIR))
def test_encode_decode_files(test_file):
    test_file = os.path.join(TEST_FILES_DIR, test_file)
    _do_test(test_file, _assert_same)


@pytest.mark.long
@pytest.mark.parametrize('test_file', os.listdir(TEST_FILES_DIR_LONG))
def test_encode_decode_files_long(test_file):
    test_file = os.path.join(TEST_FILES_DIR_LONG, test_file)
    _do_test(test_file, _assert_same)


@pytest.mark.parametrize('test_file', os.listdir(TEST_FILES_DIR_FAIL))
def test_encode_decode_files_fail(test_file):
    test_file = os.path.join(TEST_FILES_DIR_FAIL, test_file)
    with pytest.raises(SLTPParsingError):
        _do_test(test_file, _assert_same)


@pytest.mark.parametrize('test_file', os.listdir(TEST_FILES_DIR_DIFF))
def test_encode_decode_files_diff(test_file):
    test_file = os.path.join(TEST_FILES_DIR_DIFF, test_file)
    _do_test(test_file, _assert_different)



# coding=utf-8
"""
Configuration test specific to Mission & Miz
"""

import copy
from pathlib import Path

import pytest

from emiz import Mission, Miz

TEST_FILES_FOLDER = Path('./test/test_files').absolute()
if not TEST_FILES_FOLDER.exists():
    raise RuntimeError('cannot find test files')


@pytest.fixture(
    params=[
        Path(TEST_FILES_FOLDER.joinpath('TRG_KA50.miz')),
        Path(TEST_FILES_FOLDER.joinpath('test_158.miz')),
    ],
    ids=['older version', '1.5.8']
)
def test_file(request):
    yield request.param


@pytest.fixture()
def out_file():
    yield Path(TEST_FILES_FOLDER.joinpath('TRG_KA50_EMFT.miz'))


@pytest.fixture()
def bad_zip_file():
    yield Path(TEST_FILES_FOLDER.joinpath('bad_zip_file.miz'))


@pytest.fixture()
def missing_file():
    yield Path(TEST_FILES_FOLDER.joinpath('missing_files.miz'))


@pytest.fixture()
def duplicate_group_id():
    yield Path(TEST_FILES_FOLDER.joinpath('duplicate_group_id.miz'))


@pytest.fixture()
def all_objects():
    yield Path(TEST_FILES_FOLDER.joinpath('all_objects.miz'))


@pytest.fixture()
def large_file():
    yield Path(TEST_FILES_FOLDER.joinpath('TRMT_2.4.0.miz'))


@pytest.fixture()
def radio_file():
    yield Path(TEST_FILES_FOLDER.joinpath('radios.miz'))


@pytest.fixture()
def bad_files():
    yield ['bad_zip_file.miz', 'missing_files.miz']


@pytest.fixture(autouse=True)
def remove_out_file():
    """Removes OUT_FILE between tests"""
    path = Path(TEST_FILES_FOLDER.joinpath('TRG_KA50_EMFT.miz'))
    if path.exists():
        path.unlink()
    yield
    if path.exists():
        path.unlink()


with Miz(Path(TEST_FILES_FOLDER.joinpath('test_158.miz'))) as miz:
    DUMMY_MISSION = miz.mission


@pytest.fixture(scope='function')
def mission():
    yield Mission(copy.deepcopy(DUMMY_MISSION.d), copy.deepcopy(DUMMY_MISSION.l10n))


@pytest.fixture()
def mission_dict():
    yield copy.deepcopy(DUMMY_MISSION.d)


@pytest.fixture()
def mission_l10n():
    yield copy.deepcopy(DUMMY_MISSION.l10n)



# coding=utf-8
# pylint: skip-file
# FIXME: Pylint
"""Simple Lua Python Parser"""
import re

import mpmath
from natsort import natsorted

from emiz import MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)


# noinspection SpellCheckingInspection
ERRORS = {
    'unexp_type_str': 'decoding error: string expected',
    'unexp_end_string': 'Unexpected end of string while parsing Lua string.',
    'unexp_end_table': 'Unexpected end of table while parsing Lua string.',
    'mfnumber_minus': 'Malformed number (no digits after initial minus).',
    'mfnumber_dec_point': 'Malformed number (no digits after decimal point).',
    'mfnumber_sci': 'Malformed number (bad scientific format).',
}


class BaseSLTPError(Exception):
    """Base exception for SLTP module"""

    def __init__(self, *args):
        super().__init__(*args)


class SLTPParsingError(BaseSLTPError):
    """Error during parsing"""

    def __init__(self, *args):
        super().__init__(*args)


class SLTPEmptyObjectError(BaseSLTPError):
    """Error during parsing"""

    def __init__(self, *args):
        super().__init__(*args)


class SLTP:
    """Simple Lua Python Parser"""

    def __init__(self):
        LOGGER.debug('instantiating parser')
        self.text = ''
        self.ch = ''
        self.at = 0
        self.len = 0
        self.depth = 0
        self.qual = None
        self.space = re.compile(r'\s', re.M)
        self.alnum = re.compile(r'\w', re.M)
        self.line_end = re.compile(r'^(?P<intro>[ \t]*\})(?P<comment> -- end of \[.*\]),$')
        self.newline = '\n'
        self.tab = '\t'
        self.tab = '    '

    def decode(self, text):
        """Decode a Lua string to an dictionary
        :type text: str
        :rtype: dict
        :param text: string to decode
        :return: dictionary
        """
        LOGGER.debug('decoding text to dictionary')

        if not text or type(text) is not str:
            raise SLTPParsingError(ERRORS['unexp_type_str'])

        LOGGER.debug('extracting qualifier')
        qual = re.compile(r'^(?P<value>(dictionary|mission|mapResource) = ?)\n')
        match = qual.match(text)

        if match is None:
            raise ValueError('qualifier not found; first line: {}'.format(text.split('\n')[0]))

        self.qual = match.group('value')
        text = qual.sub('', text)

        reg = re.compile(r' -- .*[^\\]$', re.M)
        text = reg.sub('', text)

        self.text = text
        self.at, self.ch, self.depth = 0, '', 0
        self.len = len(text)
        self.next_chr()
        result = self.value()
        return result, self.qual

    def encode(self, obj, qualifier: str):
        """Encodes a dictionary-like object to a Lua string
        :param qualifier:
        :param obj: object to encode
        :return: valid Lua string
        """
        LOGGER.debug('encoding dictionary to text')
        if not obj:
            if qualifier.replace('=', '').rstrip() == 'mapResource':
                # Accept empty mapResource
                return '{}\n{{\n}} -- end of {}\n'.format(qualifier, qualifier.replace('=', '').rstrip())
            else:
                LOGGER.error('{}\n{{\n}} -- end of {}\n'.format(qualifier, qualifier.replace('=', '').rstrip()))
                raise SLTPEmptyObjectError(qualifier)
                # return '{}\n{{\n}} -- end of {}\n'.format(qualifier, qualifier.replace('=', '').rstrip())
        self.depth = 0
        out = []
        s = self.__encode(obj)
        lines = s.split(self.newline)
        for line in lines:
            m = self.line_end.match(line)
            if m:
                out.append('{},{}'.format(m.group('intro'), m.group('comment')))
            else:
                out.append(line)
        return '{}{} -- end of {}\n'.format(qualifier, self.newline.join(out), qualifier.replace('=', '').rstrip())

    def __encode(self, obj, dict_name=None):  # noqa C901
        s = ''
        tab = self.tab
        newline = self.newline
        if isinstance(obj, str):
            s += '"%s"' % obj.replace('"', '\\"')
        elif isinstance(obj, bool):
            s += str(obj).lower()
        elif isinstance(obj, (int, float, complex, mpmath.mpf)):
            s += str(obj)
        elif isinstance(obj, (list, tuple, dict)):
            # Ladies and gentlemen, please take a minute to behold the following code.
            # Breathe slowly, let it sink in for a minute.
            # Yes, I agree, this sucks *so* much I should hang for it.
            # Now, deal with it ^^
            self.depth += 1
            # noinspection PyTypeChecker
            if not isinstance(obj, dict) and len(filter(
                lambda x: isinstance(x, (int, float, mpmath.mpf)) or (isinstance(x, str) and len(x) < 10), obj
            )) == len(obj):
                newline = tab = ''
            dp = tab * self.depth
            s += '%s%s{%s' % (newline, tab * (self.depth - 1), newline if len(obj) > 0 else '')
            if isinstance(obj, dict):
                s += (',%s' % newline).join(
                    [dp + '[{}] ={}{}'.format(k,
                                              '' if isinstance(obj[k], (list, tuple, dict)) else ' ',
                                              self.__encode(obj[k], k)) if type(
                        k) is int else dp + '["{}"] ={}{}'.format(k, '' if isinstance(obj[k],
                                                                                      (list, tuple, dict)) else ' ',
                                                                  self.__encode(obj[k], k))
                     for k in natsorted(obj.keys(), key=str)])
            else:
                s += (',%s' % newline).join([dp + self.__encode(el) for el in obj])
            self.depth -= 1
            if len(obj) > 0:
                s += ','
            s += '%s%s}' % (newline, tab * self.depth)
        if isinstance(obj, dict):
            try:
                int(dict_name)
                s += ' -- end of [{}]'.format(dict_name)
            except (ValueError, TypeError):
                if dict_name is not None:
                    s += ' -- end of ["{}"]'.format(dict_name)
        return s

    # noinspection PyMissingOrEmptyDocstring
    def white(self):
        while self.ch:
            if self.space.match(self.ch):
                self.next_chr()
            else:
                break

    # noinspection PyMissingOrEmptyDocstring
    def next_chr(self):
        if self.at >= self.len:
            self.ch = None
            return None
        self.ch = self.text[self.at]
        self.at += 1
        return True

    # noinspection PyMissingOrEmptyDocstring
    def value(self):
        self.white()
        if not self.ch:
            return
        if self.ch == '{':
            o = self.object()
            ret = dict()
            for k in natsorted(o.keys()):
                ret[k] = o[k]
            return ret
        if self.ch == '[':
            self.next_chr()
        if self.ch in ['"', "'", '[']:
            return self.string(self.ch)
        if self.ch.isdigit() or self.ch == '-':
            return self.number()
        return self.word()

    # noinspection PyMissingOrEmptyDocstring
    def string(self, end=None):
        s = ''
        start = self.ch
        if end == '[':
            end = ']'
        if start in ['"', "'", '[']:
            while self.next_chr():
                if self.ch == end:
                    self.next_chr()
                    if start != '[' or self.ch == ']':
                        return s
                if self.ch == '\\' and start == end:
                    self.next_chr()
                    if self.ch != end:
                        s += '\\'
                s += self.ch
        raise SLTPParsingError(ERRORS['unexp_end_string'])

    # noinspection PyMissingOrEmptyDocstring
    def object(self):  # noqa C901
        o = dict()
        k = ''
        idx = 0
        numeric_keys = False
        self.depth += 1
        self.next_chr()
        self.white()
        if self.ch and self.ch == '}':
            self.depth -= 1
            self.next_chr()
            return o  # Exit here
        else:
            while self.ch:
                self.white()
                if self.ch == '{':
                    o[idx] = self.object()
                    idx += 1
                    continue
                elif self.ch == '}':
                    self.depth -= 1
                    self.next_chr()
                    if k:
                        o[idx] = k
                    if not numeric_keys and len(
                        [
                            key for key in o if type(key) in (str, float, bool, tuple, mpmath.mpf)
                        ]
                    ) == 0:
                        ar = []
                        for key in o:
                            ar.insert(key, o[key])
                        o = ar
                    return o  # or here
                else:
                    if self.ch == ',':
                        self.next_chr()
                        continue
                    else:
                        k = self.value()
                        if self.ch == ']':
                            numeric_keys = True
                            self.next_chr()
                    self.white()
                    if self.ch == '=':
                        self.next_chr()
                        self.white()
                        o[k] = self.value()
                        idx += 1
                        k = ''
                    elif self.ch == ',':
                        self.next_chr()
                        self.white()
                        o[idx] = k
                        idx += 1
                        k = ''
        raise SLTPParsingError(ERRORS['unexp_end_table'])  # Bad exit here

    # noinspection PyMissingOrEmptyDocstring
    def word(self):
        s = ''
        if self.ch != '\n':
            s = self.ch
        while self.next_chr():
            if self.alnum.match(self.ch):
                s += self.ch
            else:
                if re.match('^true$', s, re.I):
                    return True
                elif re.match('^false$', s, re.I):
                    return False
                elif s == 'nil':
                    return None
                return str(s)

    # noinspection PyMissingOrEmptyDocstring
    def number(self):  # noqa C901
        # noinspection PyMissingOrEmptyDocstring
        def next_digit(err):
            _n = self.ch
            self.next_chr()
            if not self.ch or not self.ch.isdigit():
                raise SLTPParsingError(err)
            return _n

        n = ''
        if self.ch == '-':
            n += next_digit(ERRORS['mfnumber_minus'])
        n += self.digit()
        if n == '0' and self.ch in ['x', 'X']:
            n += self.ch
            self.next_chr()
            n += self.hex()
        else:
            if self.ch and self.ch == '.':
                n += next_digit(ERRORS['mfnumber_dec_point'])
                n += self.digit()
            if self.ch and self.ch in ['e', 'E']:
                n += self.ch
                self.next_chr()
                if not self.ch or self.ch not in ('+', '-'):
                    raise SLTPParsingError(ERRORS['mfnumber_sci'])
                n += next_digit(ERRORS['mfnumber_sci'])
                n += self.digit()
        try:
            return int(n, 0)
        except ValueError:
            return mpmath.mpf(n)

    # noinspection PyMissingOrEmptyDocstring
    def digit(self):
        n = ''
        while self.ch and self.ch.isdigit():
            n += self.ch
            self.next_chr()
        return n

    # noinspection PyMissingOrEmptyDocstring
    def hex(self):
        n = ''
        while self.ch and (self.ch in 'ABCDEFabcdef' or self.ch.isdigit()):
            n += self.ch
            self.next_chr()
        return n


sltp = SLTP()



# coding=utf-8
"""
Long tests running on large files
"""

from pathlib import Path
from time import sleep

import pytest

from emiz.miz import ENCODING, Miz


@pytest.mark.long
def test_large_decode(large_file):
    miz = Miz(large_file)
    miz.unzip()
    miz.decode()


@pytest.mark.long
def test_large_zip(large_file):
    with Miz(large_file, keep_temp_dir=True) as miz:
        out_file = miz.zip()
    assert Path(out_file).exists()
    try:
        with Miz(out_file) as miz2:
            assert miz.mission.d == miz2.mission.d
            miz.mission.weather.cloud_density = 4
            assert not miz.mission.d == miz2.mission.d
            sleep(1)
            m1 = miz.mission_file
            m2 = miz2.mission_file
            with open(m1, encoding=ENCODING) as _f:
                t1 = _f.read()
            with open(m2, encoding=ENCODING) as _f:
                t2 = _f.read()
            assert t1 == t2
    finally:
        Path(out_file).unlink()



# coding=utf-8
"""
Tests Country object
"""

import pytest

from emiz.mission import Country
from emiz.miz import Miz


def test_country_generator(mission):
    for coa in (mission.blue_coa, mission.red_coa):
        assert coa.countries
    value = 0
    for country in mission.blue_coa.countries:
        assert isinstance(country, Country)
        value += 1
    assert value in [19, 21]
    value = 0
    for country in mission.red_coa.countries:
        assert isinstance(country, Country)
        value += 1
    assert value in [10, 11]


def test_get_country_by_name(mission):
    assert isinstance(mission.blue_coa.get_country_by_name('USA'), Country)
    assert isinstance(mission.blue_coa.get_country_by_id(2), Country)
    assert isinstance(mission.blue_coa.get_country_by_name('USA'), Country)
    assert isinstance(mission.blue_coa.get_country_by_id(2), Country)
    for wrong_country_name in (1, -1, 0, False, None, True):
        with pytest.raises(ValueError, msg=wrong_country_name):
            mission.blue_coa.get_country_by_name(wrong_country_name)
    for wrong_country_id in (-1, False, None, True, 'caribou'):
        with pytest.raises(ValueError, msg=wrong_country_id):
            mission.blue_coa.get_country_by_id(wrong_country_id)
    for unknown_country_name in ('nope', 'nope too'):
        with pytest.raises(ValueError, msg=unknown_country_name):
            mission.blue_coa.get_country_by_name(unknown_country_name)
    for unknown_country_id in (150,):
        with pytest.raises(ValueError, msg=unknown_country_id):
            mission.blue_coa.get_country_by_id(unknown_country_id)


def test_get_country(mission, test_file):
    assert isinstance(mission.blue_coa.get_country_by_id(2), Country)
    assert isinstance(mission.blue_coa.get_country_by_name('USA'), Country)
    with pytest.raises(ValueError):
        mission.red_coa.get_country_by_name('USA')
    with pytest.raises(ValueError):
        mission.red_coa.get_country_by_id(2)
    with Miz(test_file) as miz:
        assert isinstance(miz.mission._blue_coa.get_country_by_name('USA'), Country)



# coding=utf-8
"""
Tests global mission functionality
"""
import pytest

from emiz.mission import BaseUnit, Coalition, Country, Group, Static
from emiz.miz import Miz


def test_bullseye(mission):
    assert (11557, 371700) == mission.red_coa.bullseye_position
    assert (-291014, 617414) == mission.blue_coa.bullseye_position


def test_ln10(mission):
    import pprint
    assert mission.l10n == {'DictKey_GroupName_102': 'New Vehicle Group #024',
                            'DictKey_GroupName_105': 'New Vehicle Group #025',
                            'DictKey_GroupName_108': 'New Vehicle Group #026',
                            'DictKey_GroupName_111': 'New Vehicle Group #027',
                            'DictKey_GroupName_114': 'New Vehicle Group #028',
                            'DictKey_GroupName_117': 'New Vehicle Group #029',
                            'DictKey_GroupName_12': 'gilles',
                            'DictKey_GroupName_120': 'New Vehicle Group #030',
                            'DictKey_GroupName_123': 'New Vehicle Group #031',
                            'DictKey_GroupName_126': 'New Vehicle Group #032',
                            'DictKey_GroupName_15': 'New Static Object',
                            'DictKey_GroupName_18': 'New Static Object #001',
                            'DictKey_GroupName_21': 'New Static Object #002',
                            'DictKey_GroupName_24': 'New Static Object #003',
                            'DictKey_GroupName_27': 'New Static Object #004',
                            'DictKey_GroupName_30': 'New Vehicle Group',
                            'DictKey_GroupName_33': 'New Vehicle Group #001',
                            'DictKey_GroupName_36': 'New Vehicle Group #002',
                            'DictKey_GroupName_39': 'New Vehicle Group #003',
                            'DictKey_GroupName_42': 'New Vehicle Group #004',
                            'DictKey_GroupName_45': 'New Vehicle Group #005',
                            'DictKey_GroupName_48': 'New Vehicle Group #006',
                            'DictKey_GroupName_51': 'New Vehicle Group #007',
                            'DictKey_GroupName_54': 'New Vehicle Group #008',
                            'DictKey_GroupName_57': 'New Vehicle Group #009',
                            'DictKey_GroupName_6': 'etcher',
                            'DictKey_GroupName_60': 'New Vehicle Group #010',
                            'DictKey_GroupName_63': 'New Vehicle Group #011',
                            'DictKey_GroupName_66': 'New Vehicle Group #012',
                            'DictKey_GroupName_69': 'New Vehicle Group #013',
                            'DictKey_GroupName_72': 'New Vehicle Group #014',
                            'DictKey_GroupName_75': 'New Vehicle Group #015',
                            'DictKey_GroupName_78': 'New Vehicle Group #016',
                            'DictKey_GroupName_81': 'New Vehicle Group #017',
                            'DictKey_GroupName_84': 'New Vehicle Group #018',
                            'DictKey_GroupName_87': 'New Vehicle Group #019',
                            'DictKey_GroupName_9': 'gal',
                            'DictKey_GroupName_90': 'New Vehicle Group #020',
                            'DictKey_GroupName_93': 'New Vehicle Group #021',
                            'DictKey_GroupName_96': 'New Vehicle Group #022',
                            'DictKey_GroupName_99': 'New Vehicle Group #023',
                            'DictKey_UnitName_10': 'gal',
                            'DictKey_UnitName_100': 'Unit #023',
                            'DictKey_UnitName_103': 'Unit #024',
                            'DictKey_UnitName_106': 'Unit #025',
                            'DictKey_UnitName_109': 'Unit #026',
                            'DictKey_UnitName_112': 'Unit #027',
                            'DictKey_UnitName_115': 'Unit #028',
                            'DictKey_UnitName_118': 'Unit #029',
                            'DictKey_UnitName_121': 'Unit #030',
                            'DictKey_UnitName_124': 'Unit #031',
                            'DictKey_UnitName_127': 'Unit #032',
                            'DictKey_UnitName_13': 'gilles',
                            'DictKey_UnitName_16': '',
                            'DictKey_UnitName_19': '',
                            'DictKey_UnitName_22': '',
                            'DictKey_UnitName_25': '',
                            'DictKey_UnitName_28': '',
                            'DictKey_UnitName_31': 'Unit #1',
                            'DictKey_UnitName_34': 'Unit #001',
                            'DictKey_UnitName_37': 'Unit #002',
                            'DictKey_UnitName_40': 'Unit #003',
                            'DictKey_UnitName_43': 'Unit #004',
                            'DictKey_UnitName_46': 'Unit #005',
                            'DictKey_UnitName_49': 'Unit #006',
                            'DictKey_UnitName_52': 'Unit #007',
                            'DictKey_UnitName_55': 'Unit #008',
                            'DictKey_UnitName_58': 'Unit #009',
                            'DictKey_UnitName_61': 'Unit #010',
                            'DictKey_UnitName_64': 'Unit #011',
                            'DictKey_UnitName_67': 'Unit #012',
                            'DictKey_UnitName_7': 'etcher',
                            'DictKey_UnitName_70': 'Unit #013',
                            'DictKey_UnitName_73': 'Unit #014',
                            'DictKey_UnitName_76': 'Unit #015',
                            'DictKey_UnitName_79': 'Unit #016',
                            'DictKey_UnitName_82': 'Unit #017',
                            'DictKey_UnitName_85': 'Unit #018',
                            'DictKey_UnitName_88': 'Unit #019',
                            'DictKey_UnitName_91': 'Unit #020',
                            'DictKey_UnitName_94': 'Unit #021',
                            'DictKey_UnitName_97': 'Unit #022',
                            'DictKey_WptName_101': '',
                            'DictKey_WptName_104': '',
                            'DictKey_WptName_107': '',
                            'DictKey_WptName_11': '',
                            'DictKey_WptName_110': '',
                            'DictKey_WptName_113': '',
                            'DictKey_WptName_116': '',
                            'DictKey_WptName_119': '',
                            'DictKey_WptName_122': '',
                            'DictKey_WptName_125': '',
                            'DictKey_WptName_128': '',
                            'DictKey_WptName_14': '',
                            'DictKey_WptName_17': '',
                            'DictKey_WptName_20': '',
                            'DictKey_WptName_23': '',
                            'DictKey_WptName_26': '',
                            'DictKey_WptName_29': '',
                            'DictKey_WptName_32': '',
                            'DictKey_WptName_35': '',
                            'DictKey_WptName_38': '',
                            'DictKey_WptName_41': '',
                            'DictKey_WptName_44': '',
                            'DictKey_WptName_47': '',
                            'DictKey_WptName_50': '',
                            'DictKey_WptName_53': '',
                            'DictKey_WptName_56': '',
                            'DictKey_WptName_59': '',
                            'DictKey_WptName_62': '',
                            'DictKey_WptName_65': '',
                            'DictKey_WptName_68': '',
                            'DictKey_WptName_71': '',
                            'DictKey_WptName_74': '',
                            'DictKey_WptName_77': '',
                            'DictKey_WptName_8': '',
                            'DictKey_WptName_80': '',
                            'DictKey_WptName_83': '',
                            'DictKey_WptName_86': '',
                            'DictKey_WptName_89': '',
                            'DictKey_WptName_92': '',
                            'DictKey_WptName_95': '',
                            'DictKey_WptName_98': '',
                            'DictKey_descriptionBlueTask_3': '',
                            'DictKey_descriptionNeutralsTask_4': '',
                            'DictKey_descriptionRedTask_2': '',
                            'DictKey_descriptionText_1': '',
                            'DictKey_sortie_5': 'sortie_test'}, pprint.pprint(mission.l10n)


def test_name(mission):
    assert 'blue' == mission.blue_coa.coalition_name
    assert 'red' == mission.red_coa.coalition_name


def test_ground_control(mission):
    assert not mission.ground_control.pilots_control_vehicles
    mission.ground_control.pilots_control_vehicles = True
    assert mission.ground_control.pilots_control_vehicles
    for attrib in [
        'artillery_commander_blue',
        'artillery_commander_red',
        'forward_observer_blue',
        'forward_observer_red',
        'instructor_blue',
        'instructor_red',
        'observer_red',
        'observer_blue',
    ]:
        assert getattr(mission.ground_control, attrib) == 0
        for wrong_param in [-1, 101, True, None, 'caribou']:
            with pytest.raises(ValueError, msg=wrong_param):
                setattr(mission.ground_control, attrib, wrong_param)
        for i in range(0, 60, 5):
            setattr(mission.ground_control, attrib, i)


def test_objects(all_objects):
    with Miz(all_objects) as miz:
        _ = miz.mission._blue_coa.get_country_by_id(2)
        country = miz.mission._blue_coa.get_country_by_id(2)
        assert isinstance(country, Country)
        assert country.groups
        assert country.units
        for category in ('helicopter', 'ship', 'vehicle', 'plane'):
            assert country.get_groups_from_category(category)
            assert country.get_units_from_category(category)
            for group in country.get_groups_from_category(category):
                assert isinstance(group, Group)
            for unit in country.get_units_from_category(category):
                assert isinstance(unit, BaseUnit)
        for wrong_category in (True, -1, 0, 1, False, None, 'caribou'):
            with pytest.raises(ValueError, msg=wrong_category):
                for _ in country.get_groups_from_category(wrong_category):
                    pass
            with pytest.raises(ValueError, msg=wrong_category):
                for _ in country.get_units_from_category(wrong_category):
                    pass
        for id_ in (1, 2, 3, 4):
            assert isinstance(country.get_group_by_id(id_), Group)
            assert isinstance(country.get_unit_by_id(id_), BaseUnit)
        assert country.get_unit_by_id(5) is None
        assert country.get_group_by_id(5) is None
        for group_name in ('New Vehicle Group', 'New Helicopter Group', 'New Airplane Group', 'New Ship Group'):
            assert isinstance(country.get_group_by_name(group_name), Group)
        assert country.get_group_by_name('some other group') is None
        for unit_name in ('Unit #1', 'Unit #001', 'Pilot #001', 'Pilot #002'):
            assert isinstance(country.get_unit_by_name(unit_name), BaseUnit)
        assert country.get_unit_by_name('some other unit') is None


def test_coalitions_generator(mission):
    assert mission.coalitions
    value = 0
    for coa in mission.coalitions:
        assert isinstance(coa, Coalition)
        value += 1
    assert value == 2


def test_repr(mission):
    assert 'Mission({})'.format(mission.d) == mission.__repr__()


def test_next_group_id(mission, duplicate_group_id):
    assert mission.next_group_id == 42
    with pytest.raises(IndexError):
        with Miz(duplicate_group_id) as miz:
            assert miz.mission.next_group_id


def test_next_unit_id(mission, duplicate_group_id):
    assert mission.next_unit_id == 42
    with pytest.raises(IndexError):
        with Miz(duplicate_group_id) as miz:
            assert miz.mission.next_unit_id


def test_sortie_name(test_file, out_file):
    with Miz(test_file) as miz:
        assert miz.mission.sortie_name == 'sortie_test'
        wrong_sortie_names = [1, 0, -1, True, None]
        for wrong_sortie_name in wrong_sortie_names:
            with pytest.raises(ValueError):
                miz.mission.sortie_name = wrong_sortie_name
        miz.mission.sortie_name = 'caribou'
        assert miz.mission.sortie_name == 'caribou'
        miz.zip(out_file)
    with Miz(out_file) as miz:
        assert miz.mission.sortie_name == 'caribou'


def test_farps(mission):
    assert len(list(mission.farps())) > 0
    for farp in mission.farps():
        assert isinstance(farp, Static)



# coding=utf-8
"""
Manage mission weather
"""

import random
import typing
from datetime import date, datetime

from metar.Metar import Metar

from emiz import MAIN_LOGGER

from ..utils import hpa_to_mmhg

LOGGER = MAIN_LOGGER.getChild(__name__)

Y = 2000

SEASONS = [
    ('winter', 5, (date(Y, 1, 1), date(Y, 3, 20))),
    ('spring', 10, (date(Y, 3, 21), date(Y, 6, 20))),
    ('summer', 20, (date(Y, 6, 21), date(Y, 9, 22))),
    ('autumn', 10, (date(Y, 9, 23), date(Y, 12, 20))),
    ('winter', 5, (date(Y, 12, 21), date(Y, 12, 31))),
]

SKY_COVER = {
    "SKC": (0, 0),
    "CLR": (0, 0),
    "NSC": (0, 0),
    "NCD": (0, 0),
    "FEW": (1, 3),
    "SCT": (4, 6),
    "BKN": (7, 8),
    "OVC": (9, 10),
    "///": (0, 0),
    "VV": (0, 0)
}


def _get_season() -> typing.Tuple[str, int]:
    """
    Finds the current season based on now() and gives a dummy temperature

    Returns: tuple of season name, temperature

    """
    now = datetime.now().date().replace(year=Y)
    return next((season, temp) for season, temp, (start, end) in SEASONS if start <= now <= end)


class MissionWeather:  # pylint: disable=too-many-instance-attributes
    """
    Struct for Mission weather
    """

    def __init__(
            self,
            metar: Metar,
            min_wind: int = 0,
            max_wind: int = 40,
    ):
        self.metar = metar
        self.wind_dir = None
        self.wind_speed = None
        self._min_wind = min_wind
        self._max_wind = max_wind
        self.fog_vis = None
        self.precipitations = 0
        self.cloud_density = 0
        self.cloud_base = 300
        self.cloud_thickness = 200
        self.force_cloud_density = 0
        self.force_temperature = 999
        self._parse_precipitations()
        self._parse_clouds()

    @staticmethod
    def _random_direction() -> int:
        return random.randint(0, 359)

    @staticmethod
    def reverse_direction(heading: int) -> int:
        """
        Inverts a direction in degrees

        Args:
            heading: original direction

        Returns: direction plus or minus 180°

        """
        if heading >= 180:
            return int(heading - 180)

        return int(heading + 180)

    @staticmethod
    def _normalize_direction(heading: int) -> int:
        """
        Make sure that 0 < heading < 360

        Args:
            heading: base heading

        Returns: corrected heading

        """
        while heading > 359:
            heading = int(heading - 359)
        while heading < 0:
            heading = int(heading + 359)
        return heading

    @staticmethod
    def _gauss(mean: int, sigma: int) -> int:
        """
        Creates a variation from a base value

        Args:
            mean: base value
            sigma: gaussian sigma

        Returns: random value

        """
        return int(random.gauss(mean, sigma))

    @staticmethod
    def _deviate_wind_speed(base_speed: int, sigma: int = None) -> int:
        """
        Creates a variation in wind speed

        Args:
            base_speed: base wind speed
            sigma: sigma value for gaussian variation

        Returns: random wind speed

        """
        if sigma is None:
            sigma = base_speed / 4
        val = MissionWeather._gauss(base_speed, sigma)
        if val < 0:
            return 0
        return min(val, 50)

    @staticmethod
    def _deviate_direction(base_heading, sigma) -> int:
        """
        Creates a variation in direction

        Args:
            base_heading: base direction
            sigma: sigma value for gaussian variation

        Returns: random direction

        """
        val = MissionWeather._gauss(base_heading, sigma)
        val = MissionWeather._normalize_direction(val)
        return val

    @property
    def wind_at_ground_level_dir(self) -> int:
        """

        Returns: direction of wind at ground level

        """
        if self.metar.wind_dir is None:
            LOGGER.info('wind is variable, making a random value')
            val = self._random_direction()
        else:
            val = self.reverse_direction(self.metar.wind_dir.value())
        self.wind_dir = val
        return val

    @property
    def wind_at_ground_level_speed(self) -> int:
        """

        Returns: speed of wind at ground level

        """
        if self.metar.wind_speed is None:
            LOGGER.info('wind speed is missing, making a random value')
            val = random.randint(self._min_wind, self._max_wind)
        else:
            val = int(self.metar.wind_speed.value('MPS'))
        self.wind_speed = val
        return val

    @property
    def qnh(self) -> int:
        """

        Returns: atmospheric pressure

        """
        if self.metar.press is None:
            LOGGER.info('QNH is missing, returning standard QNH: 760')
            return 760
        return hpa_to_mmhg(self.metar.press.value())

    @property
    def visibility(self) -> int:
        """

        Returns: visibility in kilometers

        """
        if self.metar.vis is None:
            LOGGER.debug('visibility is missing, returning maximum')
            return 800000
        val = int(self.metar.vis.value())
        if val < 10000:
            self.fog_vis = min(6000, val)
        return val

    def _parse_precipitations(self):
        """

        Sets precipitation and optional minimal cloud density

        """
        if 'rain' in self.metar.present_weather():
            self.precipitations = 1
            self.force_cloud_density = 5
        if 'snow' in self.metar.present_weather():
            self.precipitations = 3
            self.force_cloud_density = 5
            self.force_temperature = 0
        if 'storm' in self.metar.present_weather():
            if self.precipitations == 2:
                self.precipitations = 4
            else:
                self.precipitations = 3
            self.force_cloud_density = 9

    def _parse_clouds(self):
        """

        Sets cloud density

        """
        layers = {}
        for sky_info in self.metar.sky:
            cover, height, _ = sky_info
            if height:
                height = int(height.value('M'))
                height = max(height, 300)
                height = min(height, 5000)
                cover = random.randint(*SKY_COVER[cover])
                layers[cover] = height
        if layers:
            max_cover = max([key for key in layers])
            self.cloud_density = max_cover
            self.cloud_base = layers[max_cover]
            self.cloud_thickness = int(max_cover / 10 * 2000)

    @property
    def temperature(self):
        """

        Sets temperature

        """
        if self.metar.temp is None:
            season, temp = _get_season()
            LOGGER.debug(f'no temperature given, since it is {season}, defaulting to {temp}')
            return temp
        value = min(self.force_temperature, int(self.metar.temp.value('C')))
        return value

    @property
    def turbulence(self) -> int:
        """

        Sets turbulence from base wind value

        """
        if self.metar.wind_gust is None:
            return 0
        val = int(self.metar.wind_gust.value('MPS'))
        if self.wind_speed >= val:
            return 0
        return int(min((val - self.wind_speed) * 10, 60))

    def apply_to_miz(self, miz):
        """
        Applies weather to an opened Miz file (the mission will be mutated)

        Args:
            miz: source miz

        Returns: True

        """

        report = ['Building mission with weather:']

        miz.mission.weather.wind_at_ground_level_dir = self.wind_at_ground_level_dir
        miz.mission.weather.wind_at_ground_level_speed = self.wind_at_ground_level_speed
        miz.mission.weather.wind_at2000_dir = self._deviate_direction(self.wind_dir, 40)
        miz.mission.weather.wind_at2000_speed = self._deviate_wind_speed(5 + self.wind_speed * 2)
        miz.mission.weather.wind_at8000_dir = self._deviate_direction(self.wind_dir, 80)
        miz.mission.weather.wind_at8000_speed = self._deviate_wind_speed(10 + self.wind_speed * 3)
        miz.mission.weather.turbulence_at_ground_level = self.turbulence

        _ground = f'{miz.mission.weather.wind_at_ground_level_dir}/{miz.mission.weather.wind_at_ground_level_speed}'
        _at2000 = f'{miz.mission.weather.wind_at2000_dir}/{miz.mission.weather.wind_at2000_speed}'
        _at8000 = f'{miz.mission.weather.wind_at8000_dir}/{miz.mission.weather.wind_at8000_speed}'
        _turbulence = f'{miz.mission.weather.turbulence_at_ground_level}'

        wind = f'Wind:' \
               f'\n\tGround: {_ground}' \
               f'\n\t2000m: {_at2000}' \
               f'\n\t8000m: {_at8000}' \
               f'\n\tTurbulence: {_turbulence}'

        report.append(wind)

        miz.mission.weather.atmosphere_type = 0
        miz.mission.weather.qnh = self.qnh

        report.append(f'Atmosphere type: {miz.mission.weather.atmosphere_type}')
        report.append(f'QNH: {miz.mission.weather.qnh}')

        miz.mission.weather.visibility = self.visibility
        if self.fog_vis:
            miz.mission.weather.fog_thickness = 1000
            miz.mission.weather.fog_visibility = self.fog_vis
            miz.mission.weather.fog_enabled = True
        else:
            miz.mission.weather.fog_enabled = False
            miz.mission.weather.fog_visibility = 0
            miz.mission.weather.fog_thickness = 0

        visibility = f'Visibility: {miz.mission.weather.visibility}' \
                     f'\n\tFog: {"yes" if miz.mission.weather.fog_enabled else "no"}' \
                     f'\n\tFog thickness: {miz.mission.weather.fog_thickness}' \
                     f'\n\tFog visibility: {miz.mission.weather.fog_visibility}'

        report.append(visibility)

        miz.mission.weather.temperature = self.temperature

        report.append(f'Temperature: {self.temperature}°C')

        miz.mission.weather.cloud_density = max(self.force_cloud_density, self.cloud_density)
        miz.mission.weather.cloud_thickness = self.cloud_thickness
        miz.mission.weather.cloud_base = self.cloud_base
        miz.mission.weather.precipitations = self.precipitations

        clouds = f'Clouds:' \
                 f'\n\tClouds density: {miz.mission.weather.cloud_density}' \
                 f'\n\tClouds thickness: {miz.mission.weather.cloud_thickness}' \
                 f'\n\tClouds base: {miz.mission.weather.cloud_base}' \
                 f'\n\tPrecipitations: {miz.mission.weather.precipitations}'

        report.append(clouds)

        LOGGER.debug(f'applying weather: {report}')

        return True



# coding=utf-8
from hypothesis import strategies as st
from hypothesis import example, given

import emiz.weather


@given(heading=st.integers(min_value=0, max_value=359))
def test_reverse_direction(heading):
    val = emiz.weather.mission_weather.MissionWeather.reverse_direction(heading)
    assert 0 <= val <= 359
    assert val == heading - 180 or val == heading + 180
    assert type(val) is int


@given(heading=st.integers(min_value=-2000, max_value=2000))
def test_normalize_direction(heading):
    val = emiz.weather.mission_weather.MissionWeather._normalize_direction(heading)
    assert 0 <= val <= 359
    assert type(val) is int


@given(base_speed=st.integers(min_value=0, max_value=40))
@example(base_speed=120)
@example(base_speed=-1)
def test_deviate_wind_speed(base_speed):
    val = emiz.weather.mission_weather.MissionWeather._deviate_wind_speed(base_speed)
    assert 0 <= val <= 50
    assert type(val) is int



# coding=utf-8
"""
Test MIZ functionality
"""
import pytest

from emiz.mission import Mission
from emiz.miz import Miz


def test_init(test_file):
    Miz(test_file)
    with pytest.raises(FileNotFoundError):
        Miz('./i_do_not_exist')


def test_context(test_file):
    with Miz(test_file) as miz:
        assert isinstance(miz.mission, Mission)
        assert isinstance(miz.l10n, dict)
        assert miz.zip_content


def test_unzip(test_file):
    miz = Miz(test_file)
    miz.unzip()


def test_decode(test_file):
    miz = Miz(test_file)
    miz.unzip()
    miz.decode()


def test_zip(out_file, test_file):
    assert not out_file.exists()
    with Miz(test_file) as miz:
        miz.zip(out_file)
    assert out_file.exists()
    with Miz(out_file) as miz2:
        assert miz.mission.d == miz2.mission.d
        miz.mission.weather.cloud_density = 4
        assert not miz.mission.d == miz2.mission.d


def test_is_unzipped(test_file):
    mis = Miz(test_file)
    assert not mis.zip_content
    mis.unzip()
    assert mis.zip_content


def test_missing_file_in_miz(missing_file):
    missing = Miz(missing_file)
    with pytest.raises(FileNotFoundError):
        missing.unzip()


def test_bad_zip_file(bad_zip_file):
    from zipfile import BadZipFile
    mis = Miz(bad_zip_file)
    with pytest.raises(BadZipFile):
        mis.unzip()


def test_temp_dir_cleaning(test_file):
    mis = Miz(test_file)
    mis.unzip()
    assert mis.temp_dir.exists()
    assert mis.temp_dir.glob('*')
    mis._remove_temp_dir()
    assert not mis.temp_dir.exists()



# coding=utf-8
"""
Tests Group object
"""

import pytest

from emiz.mission import BaseUnit, Group
from emiz.miz import Miz


def test_groups(mission):
    value = 0
    for group in mission.blue_coa.groups:
        value += 1
        assert isinstance(group, Group)
    assert value == 3
    value = 0
    for group in mission.red_coa.groups:
        value += 1
        assert isinstance(group, Group)
    assert value == 33


def test_get_groups_from_category(mission, all_objects):
    for invalid_category in ('caribou', 'Plane', 'plAne', 'ships', -1, 0, 1, True, False, None):
        with pytest.raises(ValueError):
            for _ in mission.blue_coa.get_groups_from_category(invalid_category):
                pass
    with Miz(all_objects) as miz:
        for category in ('ship', 'plane', 'helicopter', 'vehicle'):
            value = 0
            for group in miz.mission._blue_coa.get_groups_from_category(category):
                assert isinstance(group, Group)
                assert group.group_category == category
                value += 1
            assert value == 1


def test_get_group_by_id(mission):
    for group_id in range(1, 3):
        assert isinstance(mission.blue_coa.get_group_by_id(group_id), Group)
    for group_id in range(9, 41):
        assert isinstance(mission.red_coa.get_group_by_id(group_id), Group)
    for wrong_group_id in (-1, False, None, True, 'caribou'):
        with pytest.raises(ValueError, msg=wrong_group_id):
            mission.red_coa.get_group_by_id(wrong_group_id)
    for non_existing_group_id in (42, 50, 100, 150, mission.next_group_id):
        assert mission.red_coa.get_group_by_id(non_existing_group_id) is None
        assert mission.blue_coa.get_group_by_id(non_existing_group_id) is None


def test_group_by_name(mission):
    for group_name in ('etcher', 'gal', 'gilles'):
        assert isinstance(mission.blue_coa.get_group_by_name(group_name), Group)
    for i in range(1, 33):
        assert isinstance(mission.red_coa.get_group_by_name('New Vehicle Group #{}'.format(str(i).zfill(3))),
                          Group)
    for wrong_group_name in (-1, 0, 1, False, True, None):
        with pytest.raises(ValueError, msg=wrong_group_name):
            mission.blue_coa.get_group_by_name(wrong_group_name)
    for non_existing_group_name in ('yup', 'yop', 'New Vehicle group #019', 'Gilles', 'etcheR'):
        assert mission.red_coa.get_group_by_name(non_existing_group_name) is None
        assert mission.blue_coa.get_group_by_name(non_existing_group_name) is None


def test_set_hidden(test_file, out_file):
    with Miz(test_file) as miz:
        group = miz.mission.get_group_by_name('etcher')
        assert isinstance(group, Group)
        assert not group.group_hidden
        group.group_hidden = True
        assert group.group_hidden
        miz.zip(out_file)
    with Miz(out_file) as miz:
        new_group = miz.mission.get_group_by_name('etcher')
        assert isinstance(new_group, Group)
        for attrib in [x for x in Group.attribs if not x == 'group_hidden']:
            assert getattr(group, attrib) == getattr(new_group, attrib)
        assert group.group_hidden


def test_get_unit(mission):
    group = mission.get_group_by_name('etcher')
    assert isinstance(group, Group)
    unit1 = group.get_unit_by_name('etcher')
    assert isinstance(unit1, BaseUnit)
    unit2 = group.get_unit_by_id(1)
    assert isinstance(unit2, BaseUnit)
    unit3 = group.get_unit_by_index(1)
    assert isinstance(unit3, BaseUnit)
    assert unit1 == unit2 == unit3


def test_get_group_by_name(mission):
    group = mission.get_group_by_name('etcher')
    assert isinstance(group, Group)
    assert group.group_id == 1
    assert mission.get_group_by_name('le_caribou_puissant') is None



# coding=utf-8
"""
Access to AVWX API

https://avwx.rest/documentation
"""

import json

import requests
import requests.adapters

from emiz import MAIN_LOGGER

from .avwx_result import AVWXResult

LOGGER = MAIN_LOGGER.getChild(__name__)


# pylint: disable=too-few-public-methods
class AVWX:
    """
    Access to AVWX API

    https://avwx.rest/documentation
    """
    s = requests.Session()
    s.mount('https://avwx.rest', requests.adapters.HTTPAdapter(max_retries=10))

    @staticmethod
    def _query(url, params: dict = None) -> AVWXResult:
        LOGGER.debug(f'querying: {url}{params}')
        req = AVWX.s.get(url, timeout=2, params=params)
        if not req.ok:
            msg = f'failed to retrieve: {url}'
            LOGGER.error(msg)
            raise ConnectionError(msg)
        LOGGER.debug('parsing data')
        orig_data = json.loads(req.content)
        new_data = {}
        LOGGER.debug('sanitizing data keys')
        for key in orig_data:
            new_key = str(key).replace('-', '')
            new_key = new_key.lower()
            new_data[new_key] = orig_data[key]
        try:
            LOGGER.debug('returning AVWXResult instance')
            return AVWXResult(**new_data)
        except TypeError:
            import pprint
            LOGGER.error(f'invalid data was:\n{pprint.pformat(orig_data)}')
            raise

    @staticmethod
    def query_icao(icao: str) -> AVWXResult:
        """
        Queries AVWX API for weather at given station

        Args:
            icao: ICAO code of the station

        Returns: AVWXResult instance

        """
        params = {
            'options': 'info,speech,summary,translate'
        }
        LOGGER.info(f'getting METAR info for ICAO: {icao}')
        return AVWX._query(f'https://avwx.rest/api/metar/{icao}', params=params)

    @staticmethod
    def metar_to_speech(metar: str) -> str:
        """
        Creates a speakable text from a METAR

        Args:
            metar: METAR string to use

        Returns: speakable METAR for TTS

        """
        LOGGER.info(f'getting speech text from METAR: {metar}')
        params = {
            'report': metar,
            'options': 'info,speech,summary'
        }
        result = AVWX._query(f'https://avwx.rest/api/parse/metar', params=params)
        speech = str(result.speech).replace('Altimeter', 'Q N H')
        LOGGER.debug(f'resulting speech: {speech}')
        return speech
